import pytest

from yadm import fields
from yadm.documents import Document, EmbeddedDocument
from yadm.testing import create_fake


class EDocOne(EmbeddedDocument):
    type = fields.StringField(default='one')
    i = fields.IntegerField()


class EDocTwo(EmbeddedDocument):
    type = fields.StringField(default='two')
    s = fields.StringField()


class Doc(Document):
    __collection__ = 'test'
    e = fields.TypedEmbeddedDocumentField(
        type_field='type',
        types={'one': EDocOne,
               'two': EDocTwo})

    l = fields.ListField(
        fields.TypedEmbeddedDocumentField(
            type_field='type',
            types={'one': EDocOne,
                   'two': EDocTwo},
        ))


@pytest.mark.parametrize('type_name, ed_class', [
    ('one', EDocOne),
    ('two', EDocTwo),
])
def test_load(db, type_name, ed_class):
    db.db.test.insert({'e': {'type': type_name}})
    doc = db(Doc).find_one()
    assert isinstance(doc.e, ed_class)


def test_load_bad_type(db):
    db.db.test.insert({'e': {'type': 'bad_type'}})
    doc = db(Doc).find_one()

    with pytest.raises(ValueError):
        doc.e


@pytest.mark.parametrize('ed_class, type_name, data', [
    (EDocOne, 'one', {'i': 1}),
    (EDocTwo, 'two', {'s': '2'}),
])
def test_save(db, ed_class, type_name, data):
    doc = Doc()
    doc.e = ed_class(**data)
    db.save(doc)

    raw = db.db.test.find_one()
    data['type'] = type_name
    assert raw['e'] == data


def test_load_list(db):
    db.db.test.insert({'l': [
        {'type': 'one', 'i': 1},
        {'type': 'two', 'i': '2'},
    ]})

    doc = db(Doc).find_one()
    assert isinstance(doc.l[0], EDocOne)
    assert isinstance(doc.l[1], EDocTwo)


def test_fake():
    doc = create_fake(Doc)
    assert isinstance(doc.e, (EDocOne, EDocTwo))

    if isinstance(doc.e, EDocOne):
        assert isinstance(doc.e.i, int)
    elif isinstance(doc.e, EDocTwo):
        assert isinstance(doc.e.s, str)
    else:
        raise AssertionError(doc.e, type(doc.e))



"""
Fields for basic data types.
"""
from random import choice

from bson import ObjectId

from yadm.fields.base import Field, DefaultMixin, pass_null
from yadm.markers import AttributeNotSet


class StaticField(Field):
    """ Field for static data.
    """
    def __init__(self, data):
        self.data = data

    def get_default(self, document):
        return self.data

    def get_if_attribute_not_set(self, document):
        raise RuntimeError("value for {} not exist in database"
                           "".format(self.__class__.__name__))

    def copy(self):
        return self.__class__(self.data)

    def get_fake(self, document, faker, depth):
        return AttributeNotSet

    def prepare_value(self, document, value):
        raise AttributeError("can't set attribute {!r}.{}"
                             "".format(document, self.name))

    def to_mongo(self, document, value):
        if value != self.data:
            raise RuntimeError("bad value for {}: {!r} != {!r}"
                               "".format(self.__class__.__name__,
                                         value, self.data))
        else:
            return self.data

    def from_mongo(self, document, value):
        if value != self.data:
            raise RuntimeError("bad value in database for {}: {!r} != {!r}"
                               "".format(self.__class__.__name__,
                                         value, self.data))
        else:
            return self.data


class SimpleField(DefaultMixin, Field):
    """ Base field for simple types.

    :param default: default value
    :param set choices: set of possible values
    """
    type = NotImplemented
    choices = None

    def __init__(self, default=AttributeNotSet, *, choices=None, **kwargs):
        if self.type is NotImplemented:
            raise NotImplementedError("Attribute 'type' is not implemented!")

        self.choices = choices

        kwargs['default'] = default
        super().__init__(**kwargs)

        if default is not AttributeNotSet:
            self._check_choices(default)

    def get_fake(self, document, faker, depth):
        if self.choices is not None:
            return choice(self.choices)
        else:
            return super().get_fake(document, faker, depth)

    @pass_null
    def prepare_value(self, document, value):
        if value is AttributeNotSet:
            return AttributeNotSet

        elif not isinstance(value, self.type):
            value = self.type(value)

        self._check_choices(value)
        return value

    @pass_null
    def from_mongo(self, document, value):
        if value is AttributeNotSet:
            return AttributeNotSet

        elif not isinstance(value, self.type):
            value = self.type(value)

        return value

    def _check_choices(self, value):
        if self.choices is not None and value not in self.choices:
            raise ValueError("{!r} not in choices: {!r}"
                             "".format(value, self.choices))


class ObjectIdField(SimpleField):
    """ Field for ObjectId.

    :param bool default_gen: generate default value if not set
    """
    type = ObjectId
    default_gen = False

    def __init__(self, default_gen=False):
        super().__init__()
        self.default_gen = default_gen

    def get_default(self, document):
        if self.default_gen:
            return ObjectId()
        else:
            return AttributeNotSet

    def get_fake(self, document, faker, depth):
        return ObjectId()

    def copy(self):
        return self.__class__(default_gen=self.default_gen)


class BooleanField(SimpleField):
    """ Field for boolean values.
    """
    type = bool

    def get_fake(self, document, faker, depth):
        return faker.pybool()


class IntegerField(SimpleField):
    """ Field for integer.
    """
    type = int

    def get_fake(self, document, faker, depth):
        if self.choices is not None:
            return choice(self.choices)
        else:
            return faker.pyint()


class FloatField(SimpleField):
    """ Field for float.
    """
    type = float

    def get_fake(self, document, faker, depth):
        if self.choices is not None:
            return choice(self.choices)
        else:
            return faker.pyfloat()


class StringField(SimpleField):
    """ Field for string.
    """
    type = str

    def get_fake(self, document, faker, depth):
        if self.choices is not None:
            return choice(self.choices)

        try:
            fake = getattr(faker, self.name)()
        except AttributeError:
            fake = None

        if isinstance(fake, str):
            return fake
        else:
            return faker.pystr()



from yadm import fields
from yadm.documents import Document, EmbeddedDocument


class EDoc(EmbeddedDocument):
    i = fields.IntegerField()


class Doc(Document):
    __collection__ = 'testdocs'
    li = fields.SetField(fields.EmbeddedDocumentField(EDoc))


def test_save(db):
    doc = Doc()

    edoc = EDoc()
    edoc.i = 13
    doc.li.add(edoc)

    edoc = EDoc()
    edoc.i = 42
    doc.li.add(edoc)

    db.insert(doc)

    data = db.db.testdocs.find_one()

    assert 'li' in data
    assert len(data['li']) == 2
    assert 'i' in data['li'][0]

    res = set()

    for item in data['li']:
        assert 'i' in item
        assert isinstance(item['i'], int)
        res.add(item['i'])

    assert res == {13, 42}


def test_load(db):
    _id = db.db.testdocs.insert({'li': []})
    db.db.testdocs.update({'_id': _id}, {'$addToSet': {'li': {'i': 13}}})
    db.db.testdocs.update({'_id': _id}, {'$addToSet': {'li': {'i': 42}}})

    doc = db.get_queryset(Doc).find_one()

    assert hasattr(doc, 'li')
    assert len(doc.li) == 2

    res = set()

    for item in doc.li:
        assert isinstance(item, EDoc)
        assert hasattr(item, 'i')
        assert isinstance(item.i, int)
        res.add(item.i)

    assert res == {13, 42}


def test_add_to_set(db):
    doc = Doc()
    db.insert(doc)

    edoc = EDoc()
    edoc.i = 13
    doc.li.add_to_set(edoc)

    edoc = EDoc()
    edoc.i = 42
    doc.li.add_to_set(edoc)

    data = db.db.testdocs.find_one()

    assert 'li' in data
    assert len(data['li']) == 2
    assert 'i' in data['li'][0]
    assert data['li'][0]['i'] == 13
    assert data['li'][1]['i'] == 42



import abc


class Database(metaclass=abc.ABCMeta):
    """ Database.

    :param pymongo.Client client: database connection
    :param str name: database name
    :param bool read_preference: setup read_preference
    """

    @abc.abstractmethod
    def __init__(self, client, name, *, read_preference=None):
        pass

    @abc.abstractmethod
    def insert(self, document):
        """ Insert document to database.

        :param Document document: document instance for insert to database

        It's bind new document to database set
        :py:attr:`_id <yadm.documents.Document._id>`.
        """

    @abc.abstractmethod
    def save(self, document, full=False, upsert=False):
        """ Save document to database.

        :param Document document: document instance for save
        :param bool full: fully resave document
            (default: `False`)
        :param bool upsert: see documentation for MongoDB's `update`
            (default: `False`)

        If document has no `_id`
        :py:meth:`insert <Database.insert>` new document.
        """

    @abc.abstractmethod
    def update_one(self, document, reload=True, *,
                   set=None, unset=None, inc=None,
                   push=None, pull=None):  # TODO: extend
        """ Update one document.

        :param Document document: document instance for update
        :param bool reload: if True, reload document
        """

    @abc.abstractmethod
    def remove(self, document):
        """ Remove document from database.

        :param Document document: instance for remove from database
        """

    @abc.abstractmethod
    def reload(self, document, new_instance=False):
        """ Reload document.

        :param Document document: instance for reload
        :param bool new_instance: if `True` return new instance of document,
            else change data in given document (default: `False`)
        """

    @abc.abstractmethod
    def get_queryset(self, document_class, *, cache=None):
        """ Return queryset for document class.

        :param document_class: :class:`yadm.documents.Document`
        :param cache: cache for share with other querysets

        This create instance of :class:`yadm.queryset.QuerySet`
        with presetted document's collection information.
        """

    @abc.abstractmethod
    def aggregate(self, document_class, *, pipeline=None):
        """ Return aggregator for use aggregation framework.

        :param document_class: :class:`yadm.documents.Document`
        :param list pipeline: initial pipeline
        """

    @abc.abstractmethod
    def bulk(self, document_class, ordered=False, raise_on_errors=True):
        """ Return Bulk.

        :param MetaDocument document_class: class of documents fo bulk
        :param bool ordered: create ordered bulk (default `False`)
        :param bool raise_on_errors: raise BulkWriteError exception
            if write errors (default `True`)

        Context manager:

            with db.bulk(Doc) as bulk:
                bulk.insert(doc_1)
                bulk.insert(doc_2)
        """


class QuerySet(metaclass=abc.ABCMeta):
    """ Query builder.

    :param cache:
    :param dict criteria:
    :param dict projection:
    :param list sort:
    :param slice slice:
    :param int read_preference:
    """
    @abc.abstractmethod
    def __init__(self, db, document_class, *,
                 cache=None, criteria=None, projection=None, sort=None,
                 slice=None, read_preference=None):
        pass

    @abc.abstractmethod
    def __len__(self):
        return self.count()

    @abc.abstractmethod
    def __iter__(self):
        pass

    @abc.abstractmethod
    def __contains__(self, document):
        pass

    @abc.abstractmethod
    def __getitem__(self, item):
        pass

    @abc.abstractmethod
    def __call__(self, criteria=None, projection=None):
        return self.find(criteria=None, projection=None)

    @property
    @abc.abstractmethod
    def cache(self):
        """ Queryset cache object.
        """

    @abc.abstractmethod
    def copy(self, *, cache=None, criteria=None, projection=None,
             sort=None, slice=None, read_preference=None):
        """ Copy queryset with new parameters.

        Only keywords arguments is alowed.
        Parameters simply replaced with given arguments.

        :param cache:
        :param dict criteria:
        :param dict projection:
        :param list sort:
        :param slice slice:
        :param int read_preference:

        :return: new :class:`yadm.queryset.QuerySet` object
        """

    @abc.abstractmethod
    def find(self, criteria=None, projection=None):
        """ Return queryset copy with new criteria and projection.

        :param dict criteria: update queryset's criteria
        :param dict projection: update queryset's projection
        :return: new :class:`yadm.queryset.QuerySet`

        .. code:: python

            qs.find({'field': {'$gt': 3}}, {'field': True})
        """

    @abc.abstractmethod
    def find_one(self, criteria=None, projection=None, *, exc=None):
        """ Find and return only one document.

        :param dict criteria: update queryset's criteria
        :param dict projection: update queryset's projection
        :param Exception exc: raise given exception if not found
        :return: :class:`yadm.documents.Document` or **None**

        .. code:: python

            qs.find_one({'field': {'$gt': 3}}, {'field': True})
        """

    @abc.abstractmethod
    def update(self, update, *, multi=True, upsert=False):
        """ Update documents in queryset.

        :param dict update: update query
        :param bool multi: update all matched documents
            *(default True)*
        :param bool upsert: insert if not found
            *(default False)*
        :return: update result
        """

    @abc.abstractmethod
    def find_and_modify(self, update=None, *, upsert=False,
                        full_response=False, new=False,
                        **kwargs):
        """ Execute *$findAndModify* query.

        :param dict update: see second argument to update()
        :param bool upsert: insert if object doesn’t exist
            *(default False)*
        :param bool full_response: return the entire response
            object from the server *(default False)*
        :param new: return updated rather than original object
            *(default False)*
        :param kwargs: any other options the findAndModify
            command supports can be passed here
        :return: :class:`yadm.documents.Document` or **None**
        """

    @abc.abstractmethod
    def remove(self, *, multi=True):
        """ Remove documents in queryset.

        :param bool multi: if False, remove only first finded document
            *(default True)*
        """

    @abc.abstractmethod
    def fields_all(self):
        """ Clear projection.
        """

    @abc.abstractmethod
    def sort(self, *sort):
        """ Return queryset with sorting.

        :param tuples sort: tuples with two items:
            `('field_name', sort_order_as_int)`.

        .. code:: python

            qs.sort(('field_1', 1), ('field_2', -1))
        """

    @abc.abstractmethod
    def distinct(self, field):
        """ Distinct query.

        :param str field: field for distinct
        :return: list with result data
        """

    @abc.abstractmethod
    def count(self):
        """ Count documents in queryset.

        :return: **int**
        """

    @abc.abstractmethod
    def join(self, *field_names):
        """ Create `yadm.Join` object, join `field_names` and return it.

        :param str fiels_names: fields for join
        :return: new :class:`yadm.join.Join`

        Next algorithm for join:
            1. Get all documents from queryset;
            2. Aggegate all ids from requested fields;
            3. Make *$in* queries for get joined documents;
            4. Bind joined documents to objects from first queryset;

        `Join` object is instance of `abc.Sequence`.
        """

    @abc.abstractmethod
    def ids(self):
        """ Return documents ids from queryset.
        """



"""
Field with sets.

Similar as :py:mod:`yadm.fields.list`.
"""
from collections import abc

from yadm.fields.containers import Container
from yadm.fields.list import ListField


class Set(Container, abc.MutableSet):
    """ Container for set.
    """
    def __getitem__(self, item):
        raise TypeError("'{}' object does not support indexing"
                        "".format(self.__class__.__name__))

    def __setitem__(self, item, value):
        raise TypeError("'{}' object does not support item assignment"
                        "".format(self.__class__.__name__))

    def __delitem__(self, item):
        raise TypeError("'{}' object doesn't support item deletion"
                        "".format(self.__class__.__name__))

    def __eq__(self, other):
        if isinstance(other, set):
            return set(self) == set(other)
        else:
            return False

    def add(self, item):
        """ Append item to set.

        :param item: item for add

        This method does not save object!
        """
        item = self._prepare_item(len(self), item)
        if item not in self._data:
            self._data.append(item)

        self._set_changed()

    def discard(self, item):
        """ Remove item from the set if it is present.

        :param item: item for discard

        This method does not save object!
        """
        try:
            self._data.remove(item)
        except ValueError:
            pass

        self._set_changed()

    def remove(self, item):
        """ Remove item from set.

        :param item: item for remove

        This method does not save object!
        """
        try:
            self._data.remove(item)
        except ValueError as exc:
            raise KeyError from exc

        self._set_changed()

    def add_to_set(self, item, reload=True):
        """ Add item directly to database.

        :param item: item for `$addToSet`
        :param bool reload: automatically reload all values from database

        See `$addToSet` in MongoDB's `update`.
        """
        index = len(self)
        item = self._prepare_item(index, item)
        data = self._field.item_field.to_mongo(self.__document__, item)

        qs = self._get_queryset()
        qs.update({'$addToSet': {self.__field_name__: data}}, multi=False)
        self.add(item)

        if reload:
            self.reload()

    def pull(self, query, reload=True):
        """ Pull item from database.

        :param query: query for `$pull` on this field
        :param bool reload: automatically reload all values from database

        See `$pull` in MongoDB's `update`.
        """
        qs = self._get_queryset()
        qs.update({'$pull': {self.__field_name__: query}}, multi=False)

        if reload:
            self.reload()


class SetField(ListField):
    """ Field for set values.
    """
    container = Set



""" Currencies for Money.
"""
from collections import namedtuple
import random

from yadm.markers import AttributeNotSet
from yadm.fields.base import Field, DefaultMixin, pass_null


Currency = namedtuple('Currency', ['code', 'string', 'precision'])

DEFAULT_CURRENCIES_LIST = [
    Currency(784, 'AED', 2),
    Currency(971, 'AFN', 2),
    Currency(8, 'ALL', 2),
    Currency(51, 'AMD', 2),
    Currency(532, 'ANG', 2),
    Currency(973, 'AOA', 2),
    Currency(32, 'ARS', 2),
    Currency(36, 'AUD', 2),
    Currency(533, 'AWG', 2),
    Currency(944, 'AZN', 2),
    Currency(977, 'BAM', 2),
    Currency(52, 'BBD', 2),
    Currency(50, 'BDT', 2),
    Currency(975, 'BGN', 2),
    Currency(48, 'BHD', 3),
    Currency(108, 'BIF', 0),
    Currency(60, 'BMD', 2),
    Currency(96, 'BND', 2),
    Currency(68, 'BOB', 2),
    Currency(984, 'BOV', 2),
    Currency(986, 'BRL', 2),
    Currency(44, 'BSD', 2),
    Currency(64, 'BTN', 2),
    Currency(72, 'BWP', 2),
    Currency(974, 'BYR', 0),
    Currency(84, 'BZD', 2),
    Currency(124, 'CAD', 2),
    Currency(976, 'CDF', 2),
    Currency(947, 'CHE', 2),
    Currency(756, 'CHF', 2),
    Currency(948, 'CHW', 2),
    Currency(990, 'CLF', 4),
    Currency(152, 'CLP', 0),
    Currency(156, 'CNY', 2),
    Currency(170, 'COP', 2),
    Currency(970, 'COU', 2),
    Currency(188, 'CRC', 2),
    Currency(931, 'CUC', 2),
    Currency(192, 'CUP', 2),
    Currency(132, 'CVE', 0),
    Currency(203, 'CZK', 2),
    Currency(262, 'DJF', 0),
    Currency(208, 'DKK', 2),
    Currency(214, 'DOP', 2),
    Currency(12, 'DZD', 2),
    Currency(818, 'EGP', 2),
    Currency(232, 'ERN', 2),
    Currency(230, 'ETB', 2),
    Currency(978, 'EUR', 2),
    Currency(242, 'FJD', 2),
    Currency(238, 'FKP', 2),
    Currency(826, 'GBP', 2),
    Currency(981, 'GEL', 2),
    Currency(936, 'GHS', 2),
    Currency(292, 'GIP', 2),
    Currency(270, 'GMD', 2),
    Currency(324, 'GNF', 0),
    Currency(320, 'GTQ', 2),
    Currency(328, 'GYD', 2),
    Currency(344, 'HKD', 2),
    Currency(340, 'HNL', 2),
    Currency(191, 'HRK', 2),
    Currency(332, 'HTG', 2),
    Currency(348, 'HUF', 2),
    Currency(360, 'IDR', 2),
    Currency(376, 'ILS', 2),
    Currency(356, 'INR', 2),
    Currency(368, 'IQD', 3),
    Currency(364, 'IRR', 2),
    Currency(352, 'ISK', 0),
    Currency(388, 'JMD', 2),
    Currency(400, 'JOD', 3),
    Currency(392, 'JPY', 0),
    Currency(404, 'KES', 2),
    Currency(417, 'KGS', 2),
    Currency(116, 'KHR', 2),
    Currency(174, 'KMF', 0),
    Currency(408, 'KPW', 2),
    Currency(410, 'KRW', 0),
    Currency(414, 'KWD', 3),
    Currency(136, 'KYD', 2),
    Currency(398, 'KZT', 2),
    Currency(418, 'LAK', 2),
    Currency(422, 'LBP', 2),
    Currency(144, 'LKR', 2),
    Currency(430, 'LRD', 2),
    Currency(426, 'LSL', 2),
    Currency(434, 'LYD', 3),
    Currency(504, 'MAD', 2),
    Currency(498, 'MDL', 2),
    Currency(969, 'MGA', 1),
    Currency(807, 'MKD', 2),
    Currency(104, 'MMK', 2),
    Currency(496, 'MNT', 2),
    Currency(446, 'MOP', 2),
    Currency(478, 'MRO', 1),
    Currency(480, 'MUR', 2),
    Currency(462, 'MVR', 2),
    Currency(454, 'MWK', 2),
    Currency(484, 'MXN', 2),
    Currency(979, 'MXV', 2),
    Currency(458, 'MYR', 2),
    Currency(943, 'MZN', 2),
    Currency(516, 'NAD', 2),
    Currency(566, 'NGN', 2),
    Currency(558, 'NIO', 2),
    Currency(578, 'NOK', 2),
    Currency(524, 'NPR', 2),
    Currency(554, 'NZD', 2),
    Currency(512, 'OMR', 3),
    Currency(590, 'PAB', 2),
    Currency(604, 'PEN', 2),
    Currency(598, 'PGK', 2),
    Currency(608, 'PHP', 2),
    Currency(586, 'PKR', 2),
    Currency(985, 'PLN', 2),
    Currency(600, 'PYG', 0),
    Currency(634, 'QAR', 2),
    Currency(946, 'RON', 2),
    Currency(941, 'RSD', 2),
    Currency(643, 'RUB', 2),
    Currency(646, 'RWF', 0),
    Currency(682, 'SAR', 2),
    Currency(90, 'SBD', 2),
    Currency(690, 'SCR', 2),
    Currency(938, 'SDG', 2),
    Currency(752, 'SEK', 2),
    Currency(702, 'SGD', 2),
    Currency(654, 'SHP', 2),
    Currency(694, 'SLL', 2),
    Currency(706, 'SOS', 2),
    Currency(968, 'SRD', 2),
    Currency(728, 'SSP', 2),
    Currency(678, 'STD', 2),
    Currency(760, 'SYP', 2),
    Currency(748, 'SZL', 2),
    Currency(764, 'THB', 2),
    Currency(972, 'TJS', 2),
    Currency(934, 'TMT', 2),
    Currency(788, 'TND', 3),
    Currency(776, 'TOP', 2),
    Currency(949, 'TRY', 2),
    Currency(780, 'TTD', 2),
    Currency(901, 'TWD', 2),
    Currency(834, 'TZS', 2),
    Currency(980, 'UAH', 2),
    Currency(800, 'UGX', 0),
    Currency(840, 'USD', 2),
    Currency(997, 'USN', 2),
    Currency(998, 'USS', 2),
    Currency(940, 'UYI', 0),
    Currency(858, 'UYU', 2),
    Currency(860, 'UZS', 2),
    Currency(937, 'VEF', 2),
    Currency(704, 'VND', 0),
    Currency(548, 'VUV', 0),
    Currency(882, 'WST', 2),
    Currency(950, 'XAF', 0),
    Currency(961, 'XAG', 0),
    Currency(959, 'XAU', 0),
    Currency(955, 'XBA', 0),
    Currency(956, 'XBB', 0),
    Currency(957, 'XBC', 0),
    Currency(958, 'XBD', 0),
    Currency(951, 'XCD', 2),
    Currency(960, 'XDR', 0),
    Currency(952, 'XOF', 0),
    Currency(964, 'XPD', 0),
    Currency(953, 'XPF', 0),
    Currency(962, 'XPT', 0),
    Currency(994, 'XSU', 0),
    Currency(963, 'XTS', 0),
    Currency(965, 'XUA', 0),
    Currency(999, 'XXX', 0),
    Currency(886, 'YER', 2),
    Currency(710, 'ZAR', 2),
    Currency(967, 'ZMW', 2),
]


class CurrencyStorage(dict):
    def __init__(self, currencies):
        super().__init__()
        for code, string, precision in currencies:
            currency = Currency(code, string, precision)
            self[code] = currency
            self[string] = currency

    def __getitem__(self, item):
        if isinstance(item, (str, int)):
            if item in self:
                return super().__getitem__(item)
            else:
                raise KeyError("{!r} is invalid value of CurrencyStorage."
                               "".format(item))

        elif isinstance(item, Currency):
            return self[item.code]

        else:
            raise TypeError("{!r} is invalid type value of CurrencyStorage."
                            "".format(type(item)))


DEFAULT_CURRENCY_STORAGE = CurrencyStorage(DEFAULT_CURRENCIES_LIST)


class CurrencyField(DefaultMixin, Field):
    def __init__(self, *,
                 default=AttributeNotSet,
                 currency_storage=DEFAULT_CURRENCY_STORAGE):
        self._currency_storage = currency_storage

        if default is AttributeNotSet:
            pass
        elif isinstance(default, (Currency, str, int)):
            default = self._currency_storage[default]
        else:
            raise TypeError("Bad type for default.")

        super().__init__(default=default)

    def get_fake(self, document, faker, depth):
        return random.choice(list(DEFAULT_CURRENCY_STORAGE.values()))

    @pass_null
    def prepare_value(self, document, value):
        if isinstance(value, Currency):
            return value
        elif isinstance(value, (str, int)):
            if value in self._currency_storage:
                return self._currency_storage[value]
            else:
                raise ValueError("Invalid value for CurrencyField.")
        else:
            raise TypeError("Only Currency or None is allowed for CurrencyField.")

    @pass_null
    def to_mongo(self, document, value):
        return value.string

    @pass_null
    def from_mongo(self, document, value):
        return self._currency_storage[value]



import re
import sys
from unittest import SkipTest

import pytest
import pymongo

from yadm.database import Database


try:
    import motor  # noqa
    WITH_MOTOR = True
except ImportError:
    WITH_MOTOR = False


def pytest_addoption(parser):
    parser.addoption(
        '--mongo',
        action='store',
        default='localhost:27017/test',
        help='MongoDB connection uri',
        metavar='"localhost:27017/test"',
    )


def pytest_ignore_collect(path, config):
    if '/tests_aio' in str(path):
        return not WITH_MOTOR or sys.version_info < (3, 6)


@pytest.fixture(scope='session')
def mongo_args(request):
    uri = request.config.getoption('--mongo')
    res = re.match(r'([\w\.]+)(?::(\d+))?(?:\/(.+))?', uri)

    if not res:
        raise ValueError(uri)

    host, port, name = res.groups()

    if not host:
        host = 'localhost'

    if not port:
        port = 27017

    if not name:
        name = 'test'

    return (host, int(port), name)


@pytest.fixture(scope='session')
def client(request, mongo_args):
    host, port, name = mongo_args
    try:
        return pymongo.MongoClient(host, port, tz_aware=True)
    except pymongo.errors.ConnectionFailure:
        raise SkipTest("Can't connect to mongodb ({}:{})".format(host, port))


@pytest.fixture(scope='function')
def db(client, mongo_args):
    host, port, name = mongo_args
    client.drop_database(name)
    return Database(client, name)



import pytest
from bson import ObjectId

from yadm.documents import Document
from yadm.fields.simple import StringField, IntegerField


class Doc(Document):
    choices = StringField(choices={'qwerty', 'zzz', 'asd'})
    choices_w_default = IntegerField(default=13, choices={0, 13, 42})


def test_choices_set_valid():
    doc = Doc()
    doc.choices = 'zzz'
    assert doc.choices == 'zzz'


def test_choices_set_invalid():
    doc = Doc()

    with pytest.raises(ValueError):
        doc.choices = 'invalid'


def test_choices_w_default_valid():
    doc = Doc()
    assert doc.choices_w_default == 13


def test_choices_w_default_invalid():
    with pytest.raises(ValueError):
        IntegerField(default=1, choices={0, 13, 42})


def test_id_default():
    doc = Doc()

    with pytest.raises(AttributeError):
        doc._id

    with pytest.raises(AttributeError):
        doc.id


def test_id_set():
    _id = ObjectId()
    doc = Doc()
    doc.id = _id

    assert doc.id is _id
    assert doc.id is doc._id


def test_id_set_str():
    _id = ObjectId()
    doc = Doc()
    doc._id = str(_id)

    assert isinstance(doc._id, ObjectId)
    assert doc._id == _id


def test_repr():
    assert '_id' in repr(Doc._id)



import decimal

import pytest

from yadm import fields
from yadm.serialize import to_mongo, from_mongo
from yadm.fields.money import DEFAULT_CURRENCY_STORAGE
from yadm.documents import Document


class TestMoney:
    def test_general(self):
        assert str(fields.Money(1, 'USD')) == '1.00 USD'
        assert str(fields.Money('1.2345', 840)) == '1.24 USD'
        assert str(fields.Money(10, 'RUB')) == '10.00 RUB'
        assert str(fields.Money('99', 'CHF')) == '99.00 CHF'

        assert str(fields.Money(5, 840)) == '5.00 USD'
        assert str(fields.Money(7, 826)) == '7.00 GBP'

        r1 = fields.Money(10, 'RUB')
        r2 = fields.Money(r1)
        assert r1 == r2
        assert r1 is not r2

        with pytest.raises(TypeError):
            fields.Money(r1, 'RUB')

    def test_from_string(self):
        assert fields.Money.from_string('1.00 USD') == fields.Money(1, 'USD')
        assert fields.Money.from_string('  1.2345  USD') == fields.Money('1.2345', 'USD')
        assert fields.Money.from_string('1\tUSD\n') == fields.Money(1, 'USD')
        assert fields.Money.from_string('1 840') == fields.Money(1, 'USD')

        with pytest.raises(decimal.InvalidOperation):
            fields.Money.from_string('one USD')

        with pytest.raises(ValueError):
            fields.Money.from_string('1 A01')

    def test_arithmetic_operations(self):
        r10 = fields.Money(10, 'RUB')
        r1 = fields.Money('1', 'RUB')
        u10 = fields.Money('10', 'USD')
        u1 = fields.Money(1, 'USD')

        assert r10 + r1 == fields.Money(11, 'RUB')
        assert r1 + r10 == fields.Money(11, 'RUB')
        assert r10 - r1 == fields.Money(9, 'RUB')
        assert r1 - r10 == fields.Money(-9, 'RUB')
        assert u10 + u1 == fields.Money(11, 'USD')
        assert u1 + u10 == fields.Money(11, 'USD')
        assert u10 - u1 == fields.Money(9, 'USD')
        assert u1 - u10 == fields.Money(-9, 'USD')

        assert r10 * 10 == fields.Money(100, 'RUB')
        assert 10 * r10 == fields.Money(100, 'RUB')
        assert r10 / 10 == fields.Money(1, 'RUB')
        assert u1 * 10 == fields.Money(10, 'USD')
        assert 10 * u1 == fields.Money(10, 'USD')
        assert u1 / 10 == fields.Money('0.1', 'USD')

    def test_negative_unary_operations(self):
        u1 = fields.Money('1.01', 'USD')

        assert -u1 == fields.Money('-1.01', 'USD')
        assert -(-u1) == u1

    def test_is_methods(self):
        r0 = fields.Money(0, 'RUB')
        r1 = fields.Money(1, 'RUB')
        r_1 = fields.Money(-1, 'RUB')

        assert r0.is_zero()
        assert r1.is_positive()
        assert r_1.is_negative()

    def test_cmp(self):
        r0 = fields.Money(0, 'RUB')
        r1 = fields.Money(1, 'RUB')
        r_1 = fields.Money(-1, 'RUB')
        r10 = fields.Money(10, 'RUB')
        u1 = fields.Money(1, 'USD')

        assert r10 > r1
        assert not r10 < r1
        assert r1 < r10
        assert not r1 > r10
        assert r10 >= r1
        assert not r10 <= r1
        assert r1 <= r10
        assert not r1 >= r10

        assert r1 > 0
        assert not r1 < 0
        assert r_1 < 0
        assert not r_1 > 0

        assert not bool(r0)
        assert bool(r1)

        assert r0 == 0
        assert not r0 != 0

        with pytest.raises(ValueError) as e:
            r10 > u1

        with pytest.raises(TypeError) as e:
            r0 > 1

    def test_forbidden_operations(self):
        r1 = fields.Money('1', 'RUB')
        u1 = fields.Money(1, 'USD')

        with pytest.raises(ValueError) as e:
            u1 + r1

        s = "Money '1.00 USD' and '1.00 RUB' must be with the same currencies"
        assert s in str(e.value)

        with pytest.raises(TypeError) as e:
            u1 * r1

        s = "unsupported operand type(s) for *: 'Money' and 'Money'"
        assert s in str(e.value)

        with pytest.raises(TypeError) as e:
            u1 + 10

        s = "unsupported operand type(s) for +: 'Money' and 'int'"
        assert s in str(e.value)

        with pytest.raises(TypeError) as e:
            10 + u1

        s = "unsupported operand type(s) for +: 'int' and 'Money'"
        assert s in str(e.value)

    @pytest.mark.parametrize('iv, ov', [
        ('3.14', '3.14'),
        ('0.14', '0.14'),
        ('30.14', '30.14'),
        ('3.1', '3.10'),
        ('0.1', '0.10'),
    ])
    @pytest.mark.parametrize('cur', ['RUB', 'EUR', 'USD'])
    def test_money_str(self, iv, cur, ov):
        assert str(fields.Money(iv, cur)) == '{} {}'.format(ov, cur)

    @pytest.mark.parametrize('iv, ov', [
        ('3.14', 314),
        ('0.14', 14),
        ('30.14', 3014),
        ('3.1', 310),
        ('0.1', 10),
    ])
    @pytest.mark.parametrize('cur', ['RUB', 'EUR', 'USD'])
    @pytest.mark.parametrize('func', [
        lambda money: money.total_cents,
        lambda money: money.to_mongo()[0],  # b/c function, for coverage
    ])
    def test_total_cents(self, func, iv, cur, ov):
        assert func(fields.Money(iv, cur)) == ov

    def test_total_cents_rounding(self):
        money = fields.Money(decimal.Decimal('1.235'), 'RUB')
        assert money._context is fields.Money._context
        assert money.total_cents == 124

    @pytest.mark.parametrize('iv, ov', [
        ((1000, 'RUB'), fields.Money(10, 'RUB')),
        ((1000, 'JPY'), fields.Money(1000, 'JPY')),
        ((1000, 'MGA'), fields.Money(100, 'MGA')),
        ((1000, DEFAULT_CURRENCY_STORAGE['MRO']), fields.Money(100, 'MRO')),
    ])
    def test_from_cents(self, iv, ov):
        assert fields.Money.from_cents(iv[0], iv[1]) == ov

    def tests_abs(self):
        value = abs(fields.Money('-3.14', 'RUB'))
        assert isinstance(value, fields.Money)
        assert value.currency.string == 'RUB'
        assert value > 0
        assert value.total_cents > 0

    @pytest.mark.parametrize('first, seccond, equal', [
        (fields.Money('3.14', 'RUB'), fields.Money('3.14', 'RUB'), True),
        (fields.Money('3.14', 'RUB'), fields.Money('2.48', 'RUB'), False),
        (fields.Money('3.14', 'RUB'), fields.Money('-3.14', 'RUB'), False),
        (fields.Money('3.14', 'RUB'), fields.Money('3.14', 'EUR'), False),
    ])
    def tests_hash(self, first, seccond, equal):
        assert (hash(first) == hash(seccond)) is equal

    @pytest.mark.parametrize('first, seccond, result', [
        (fields.Money('10', 'RUB'), fields.Money('2', 'RUB'), decimal.Decimal('5.00')),
        (fields.Money('10', 'RUB'), fields.Money('0.5', 'RUB'), decimal.Decimal('20')),
        (fields.Money('2', 'RUB'), fields.Money('10', 'RUB'), decimal.Decimal('0.20')),
        (fields.Money('2', 'RUB'), 10, fields.Money('0.20', 'RUB')),
    ])
    def test_truediv(self, first, seccond, result):
            assert first / seccond == result
            assert isinstance(first / seccond, result.__class__)

    def test_truediv__error(self):
            with pytest.raises(ValueError):
                fields.Money('10', 'RUB') / fields.Money('2', 'EUR')

    @pytest.mark.parametrize('first, seccond, result', [
        (fields.Money('5', 'RUB'), 2, fields.Money('10.00', 'RUB')),
        (fields.Money('2', 'RUB'), decimal.Decimal('5.5'), fields.Money('11', 'RUB')),
        (fields.Money('10', 'RUB'), decimal.Decimal('0.3'), fields.Money('3', 'RUB')),
    ])
    def test_mul(self, first, seccond, result):
            assert first * seccond == result
            assert isinstance(first / seccond, result.__class__)

    def test_mul__error(self):
            with pytest.raises(TypeError):
                fields.Money('10', 'RUB') * 1.2

    def test_rounding(self):
        money = fields.Money(decimal.Decimal('1.233'), 'RUB')
        assert money == fields.Money(decimal.Decimal('1.24'), 'RUB')
        assert money.value == decimal.Decimal('1.24')


class TestCurrencyStorage:
    @pytest.mark.parametrize('key, code, string, precision', [
        (643, 643, 'RUB', 2),
        ('RUB', 643, 'RUB', 2),
        (840, 840, 'USD', 2),
        ('USD', 840, 'USD', 2),
        (368, 368, 'IQD', 3),
        ('IQD', 368, 'IQD', 3),
    ])
    def test_currency_storage(self, key, code, string, precision):
        assert DEFAULT_CURRENCY_STORAGE[key].code == code
        assert DEFAULT_CURRENCY_STORAGE[key].string == string
        assert DEFAULT_CURRENCY_STORAGE[key].precision == precision

    @pytest.mark.parametrize('key', ['BLABLABLA', 100500])
    def test_currency_storage__key_error(self, key):
        with pytest.raises(KeyError):
            DEFAULT_CURRENCY_STORAGE[key]


class TestMoneyField:
    class Doc(Document):
        __collection__ = 'testdocs'
        money = fields.MoneyField(bcc='RUB')

    @pytest.mark.parametrize('value, currency, v, c', [
        ('3.14', 'USD', 314, 840),
        (1, 'CLF', 10000, 990),
        (1, 'BYR', 1, 974),
    ])
    def test_save(self, db, value, currency, v, c):
        doc = self.Doc()
        doc.money = fields.Money(value, currency)
        db.save(doc)

        data = db.db.testdocs.find_one()

        assert 'money' in data
        assert isinstance(data['money'], list)
        assert data['money'] == [v, c]

    @pytest.mark.parametrize('value, currency, v, c', [
        ('3.14', 'USD', 314, 840),
        (1, 'CLF', 10000, 990),
        (1, 'BYR', 1, 974),
    ])
    def test_load(self, db, value, currency, v, c):
        db.db.testdocs.insert({'money': [v, c]})

        doc = db(self.Doc).find_one()

        assert hasattr(doc, 'money')
        assert isinstance(doc.money, fields.Money)
        assert doc.money == fields.Money(value, currency)

    def test_load_and_save(self, db):
        db.db.testdocs.insert({'money': [840, 999]})

        doc = db(self.Doc).find_one()
        doc.money = fields.Money('10.5', 'RUB')
        db.save(doc)

        assert hasattr(doc, 'money')
        assert isinstance(doc.money, fields.Money)
        assert doc.money == fields.Money('10.5', 'RUB')

        data = db.db.testdocs.find_one()

        assert 'money' in data
        assert isinstance(data['money'], list)
        assert data['money'] == [1050, 643]

    def test_load_and_save_bbc(self, db):
        db.db.testdocs.insert({'money': 1234})

        doc = db(self.Doc).find_one()

        assert hasattr(doc, 'money')
        assert isinstance(doc.money, fields.Money)
        assert doc.money == fields.Money('12.34', 'RUB')

    def test_serialize_cycle(self, db):
        doc = self.Doc()
        doc.money = fields.Money(1, 'RUB')
        db.save(doc)

        data = db(self.Doc).find_one()

        data_tm = to_mongo(data)
        assert data_tm['money'][0] == 100
        assert data_tm['money'][1] == 643

        data_fm = from_mongo(self.Doc, data_tm)
        assert data_fm.money == fields.Money(1, 'RUB')


class TestCurrencyField:
    class Doc(Document):
        __collection__ = 'testdocs'
        currency = fields.CurrencyField(default='EUR')

    def test_bad_type(self):
        doc = self.Doc()
        with pytest.raises(TypeError):
            doc.currency = object()

    @pytest.mark.parametrize('value', ['ERROR', 100500])
    def test_bad_value(self, value):
        doc = self.Doc()
        with pytest.raises(ValueError):
            doc.currency = value

    @pytest.mark.parametrize('currency', ['USD', 'RUB', 'IQD'])
    def test_save(self, db, currency):
        doc = self.Doc()
        doc.currency = DEFAULT_CURRENCY_STORAGE[currency]
        db.save(doc)

        data = db.db.testdocs.find_one()

        assert 'currency' in data
        assert isinstance(data['currency'], str)
        assert data['currency'] == currency

    @pytest.mark.parametrize('currency', ['USD', 'RUB', 'IQD'])
    def test_load(self, db, currency):
        db.db.testdocs.insert({'currency': currency})

        doc = db(self.Doc).find_one()

        assert hasattr(doc, 'currency')
        assert isinstance(doc.currency, fields.Currency)
        assert doc.currency is DEFAULT_CURRENCY_STORAGE[currency]



""" This package contain all fields.
"""

from yadm.fields.base import FieldDescriptor, Field, NotLoadedError  # noqa

from yadm.fields.simple import (  # noqa
    StaticField,
    ObjectIdField,
    BooleanField,
    IntegerField,
    FloatField,
    StringField,
)

from yadm.fields.email import EmailField  # noqa
from yadm.fields.enum import EnumField, EnumStateField  # noqa

from yadm.fields.datetime import (  # noqa
    DatetimeField,
    TimedeltaField,
)
from yadm.fields.decimal import DecimalField  # noqa
from yadm.fields.money import (  # noqa
    Money,
    MoneyField,
    Currency,
    CurrencyField,
    DEFAULT_CURRENCY_STORAGE
)

from yadm.fields.embedded import (   # noqa
    EmbeddedDocumentField,
    TypedEmbeddedDocumentField,
    SimpleEmbeddedDocumentField,
)

from yadm.fields.list import ListField  # noqa
from yadm.fields.set import SetField  # noqa
from yadm.fields.map import MapField, MapCustomKeysField  # noqa
from yadm.fields.mongo_map import UnmutableMap, MongoMapField  # noqa

from yadm.fields.reference import (  # noqa
    ReferenceField,
    BrokenReference,
    NotBindingToDatabase,
)

from yadm.fields.geo import Point, PointField  # noqa
from yadm.fields.geo import MultiPoint, MultiPointField  # noqa



import pytest

from bson import ObjectId

from yadm.documents import Document
from yadm import fields


class DocRef(Document):
    __collection__ = 'testdocs_ref'
    i = fields.IntegerField()


class Doc(Document):
    __collection__ = 'testdocs'
    i = fields.IntegerField()
    ref = fields.ReferenceField(DocRef)


@pytest.fixture
def id_ref_1(db):
    return db.db.testdocs_ref.insert({'i': 1})


@pytest.fixture
def id_ref_2(db, id_ref_1):
    return db.db.testdocs_ref.insert({'i': 2})


@pytest.fixture
def qs(db, id_ref_1, id_ref_2):
    for n in range(10):
        db.db.testdocs.insert({
            'i': n,
            'ref': id_ref_1 if n % 2 else id_ref_2,
        })

    return db(Doc)


@pytest.fixture
def join(qs):
    return qs.join('ref')


def test_len(join):
    assert len(join) == 10


def test_iter(join):
    assert len(list(join)) == 10


def test_join(join, id_ref_1, id_ref_2):
    for doc in join:
        assert 'ref' in doc.__raw__
        assert isinstance(doc.__raw__['ref'], ObjectId)
        assert 'ref' in doc.__cache__
        ref = doc.__cache__['ref']
        assert isinstance(ref, Document)
        assert doc.ref.id == id_ref_1 if doc.i % 2 else id_ref_2


def test_get_queryset(db, qs, id_ref_1, id_ref_2):
    db.db.testdocs_ref.insert({'i': 3})
    qs = qs.join().get_queryset('ref')
    assert qs.count() == 2
    assert {d.id for d in qs} == {id_ref_1, id_ref_2}



from datetime import datetime, timedelta

import pytest

import pytz

from yadm import fields
from yadm.documents import Document
from yadm.testing import create_fake


class TestDatetimeField:
    class Doc(Document):
        __collection__ = 'testdocs'
        dt = fields.DatetimeField()
        now = fields.DatetimeField(auto_now=True)

    def test_cast(self):
        doc = self.Doc()
        doc.dt = datetime(1970, 1, 1, tzinfo=pytz.utc).isoformat()
        assert isinstance(doc.dt, datetime)

    def test_default_none(self):
        doc = self.Doc()
        assert not hasattr(doc, 'dt')

    def test_save(self, db):
        doc = self.Doc()
        doc.dt = datetime(1970, 1, 1, tzinfo=pytz.utc)
        db.insert(doc)

        doc = db.get_queryset(self.Doc).find_one(doc.id)
        data = db.db.testdocs.find_one()

        assert isinstance(data['dt'], datetime)
        assert data['dt'] == doc.dt

    def test_load(self, db):
        epoch = datetime(1970, 1, 1, tzinfo=pytz.utc)
        _id = db.db.testdocs.insert({'dt': epoch})

        doc = db.get_queryset(self.Doc).find_one(_id)

        assert isinstance(doc.dt, datetime)
        assert doc.dt == epoch

    def test_default_auto_now(self):
        doc = self.Doc()
        assert hasattr(doc, 'now')
        assert isinstance(doc.now, datetime)

    def test_default_now_save(self, db):
        doc = self.Doc()
        db.insert(doc)

        doc = db.get_queryset(self.Doc).find_one(doc.id)
        data = db.db.testdocs.find_one()

        assert isinstance(data['now'], datetime)
        assert data['now'] == doc.now

    def test_now_save(self, db):
        epoch = datetime(1970, 1, 1, tzinfo=pytz.utc)

        doc = self.Doc()
        doc.now = epoch
        db.insert(doc)

        doc = db.get_queryset(self.Doc).find_one(doc.id)
        data = db.db.testdocs.find_one()

        assert isinstance(data['now'], datetime)
        assert data['now'] == epoch

    def test_now_load(self, db):
        epoch = datetime(1970, 1, 1, tzinfo=pytz.utc)
        _id = db.db.testdocs.insert({'now': epoch})

        doc = db.get_queryset(self.Doc).find_one(_id)

        assert isinstance(doc.now, datetime)
        assert doc.now == epoch


class TestTimedeltaField:
    class Doc(Document):
        __collection__ = 'testdocs'
        td = fields.TimedeltaField()
        tdd = fields.TimedeltaField(default=timedelta(seconds=13))

    def test_init(self):
        doc = self.Doc()

        assert not hasattr(doc, 'td')
        assert doc.tdd == timedelta(seconds=13)

    def test_save(self, db):
        doc = self.Doc()

        doc.td = timedelta(seconds=26)
        db.save(doc)

        raw = db.db['testdocs'].find_one()

        assert raw['td'] == 26
        assert raw['tdd'] == 13

    def test_load(self, db):
        db.db['testdocs'].insert_one({'td': 26})

        doc = db(self.Doc).find_one()

        assert doc.td == timedelta(seconds=26)
        assert not hasattr(doc, 'tdd')

    def test_set__type_error(self):
        doc = self.Doc()

        with pytest.raises(TypeError):
            doc.td = 'string is wrong type'

    def test_faker(self):
        doc = create_fake(self.Doc)

        assert isinstance(doc.td, timedelta)
        assert isinstance(doc.tdd, timedelta)



from enum import Enum

from bson import ObjectId
from pymongo import read_preferences

from yadm.join import Join
from yadm.cache import StackCache
from yadm.results import UpdateResult, RemoveResult
from yadm.serialize import from_mongo

CACHE_SIZE = 100

_Primary = read_preferences.Primary()
_PrimaryPreferred = read_preferences.PrimaryPreferred()


class NotFoundBehavior(Enum):
    NONE = 'none'
    SKIP = 'skip'
    ERROR = 'error'


class NotFoundError(Exception):
    pass


class BaseQuerySet:
    """ Query builder.

    :param db:
    :param document_class:
    :param cache:
    :param dict criteria:
    :param dict projection:
    :param list sort:
    :param slice slice:
    :param dict collection_params:
    """
    def __init__(self, db, document_class, *,
                 cache=None, criteria=None, projection=None, sort=None,
                 slice=None, batch_size=None, collection_params=None):

        self._db = db
        self._document_class = document_class
        self._cache = cache
        self._criteria = {} if criteria is None else criteria
        self._projection = projection
        self._sort = sort
        self._slice = slice
        self._batch_size = batch_size
        self._collection_params = collection_params or {}

    def __repr__(self):
        return ("{s.__class__.__name__}({s._document_class.__collection__}"
                " {s._criteria!r} {s._projection!r} {s._sort!r})"
                "".format(s=self))

    def __call__(self, criteria=None, projection=None):
        return self.find(criteria, projection)

    def __getitem__(self, item):
        if isinstance(item, slice):
            if item.step is not None:
                n = self.__class__.__name__
                raise TypeError("{} not support slicing with step".format(n))

            elif (item.start and item.start < 0) or (item.stop and item.stop < 0):
                n = self.__class__.__name__
                raise TypeError("{} not support negative slicing".format(n))

            return self.copy(slice=item)

        elif isinstance(item, int):
            return self._get_one(item)

        else:
            raise TypeError("Only slice or int accepted, but {}"
                            "".format(item.__class__))

    def _from_mongo_one(self, data, *, projection=None):
        """ Create document from raw data.
        """
        projection = projection or self._projection

        if data is None:
            return None
        elif not projection:
            not_loaded = ()
        else:
            include = [f for f, v in projection.items() if v]
            exclude = {f for f, v in projection.items() if not v}

            if include:
                if exclude and exclude != {'_id'}:
                    raise ValueError("projection cannot have a mix"
                                     " of inclusion and exclusion")

                for field_name in self._document_class.__fields__:
                    if field_name not in include:
                        exclude.add(field_name)

            not_loaded = exclude

        doc = from_mongo(self._document_class, data, not_loaded)
        doc.__db__ = self._db
        doc.__qs__ = self
        return doc

    @property
    def _collection(self):  # noqa
        """ pymongo collection.
        """
        return self._db._get_collection(self._document_class,
                                        params=self._collection_params)

    @property
    def _cursor(self):
        """ Raw cursor with parameters from queryset.
        """
        cursor = self._collection.find(self._criteria, self._projection or None)

        if self._sort is not None:
            cursor = cursor.sort(self._sort)

        if self._slice is not None:
            if self._slice.start:
                cursor = cursor.skip(self._slice.start)

            if self._slice.stop:
                cursor = cursor.limit(self._slice.stop - (self._slice.start or 0))

        if self._batch_size is not None:
            cursor = cursor.batch_size(self._batch_size)

        return cursor

    @property
    def cache(self):
        """ Queryset cache object.
        """
        if self._cache is None:
            self._cache = StackCache(size=CACHE_SIZE)

        return self._cache

    def copy(self, *, cache=None, criteria=None, projection=None,
             sort=None, slice=None, batch_size=None, collection_params=None):
        """ Copy queryset with new parameters.

        Only keywords arguments is alowed.
        Parameters simply replaced with given arguments.

        :param cache:
        :param dict criteria:
        :param dict projection:
        :param list sort:
        :param slice slice:
        :param dict collection_params:

        :return: new :class:`yadm.queryset.QuerySet` object
        """
        return self.__class__(
            self._db, self._document_class,
            cache=cache or self._cache,
            criteria=criteria or self._criteria,
            projection=projection or self._projection,
            sort=sort or self._sort,
            slice=slice or self._slice,
            batch_size=batch_size or self._batch_size,
            collection_params=collection_params or self._collection_params,
        )

    def read_preference(self, read_preference):
        """ Setup readPreference.

        Return new QuerySet instance.

        Deprecated since 1.4.0.
        Use `collection_params` argument in `copy`.
        """
        collection_params = (self._collection_params or {}).copy()
        collection_params['read_preference'] = read_preference
        return self.copy(collection_params=collection_params)

    def read_primary(self, preferred=False):
        """ Return queryset with setupd read concern for primary.

        If `preferred` argument is `True`, `PrimaryPreferred` is used
        else `Primary`.
        """
        collection_params = (self._collection_params or {}).copy()

        if not preferred:
            collection_params['read_preference'] = _Primary
        else:
            collection_params['read_preference'] = _PrimaryPreferred

        return self.copy(collection_params=collection_params)

    def find(self, criteria=None, projection=None):
        """ Return queryset copy with new criteria and projection.

        :param dict criteria: update queryset's criteria
        :param dict projection: set queryset's projection
        :return: new :class:`yadm.queryset.QuerySet`

        .. code:: python

            qs({'field': {'$gt': 3}}, {'field': True})
        """
        if criteria is not None:
            if not self._criteria:
                criteria_new = criteria
            elif set(self._criteria) & set(criteria):
                criteria_new = {'$and': [self._criteria, criteria]}
            else:
                criteria_new = self._criteria.copy()
                criteria_new.update(criteria)
        else:
            criteria_new = None

        if projection is not None:
            projection_new = projection
        elif self._projection is not None:
            projection_new = self._projection.copy()
        else:
            projection_new = None

        return self.copy(criteria=criteria_new, projection=projection_new)

    def fields(self, *fields):
        """ Get only setted fields.

        Update projection with fields.

        :param str fields:
        :return: new :class:`yadm.queryset.QuerySet`

        .. code:: python

            qs('field', 'field2')
        """
        return self.find(projection=dict.fromkeys(fields, True))

    def fields_all(self):
        """ Clear projection.
        """
        qs = self.copy()
        qs._projection = None
        return qs

    def sort(self, *sort):
        """ Return queryset with sorting.

        :param tuples sort: tuples with two items:
            `('field_name', sort_order_as_int)`.

        .. code:: python

            qs.sort(('field_1', 1), ('field_2', -1))
        """
        sort = list(sort)

        if self._sort is None:
            return self.copy(sort=sort)
        else:
            return self.copy(sort=self._sort + sort)

    def batch_size(self, batch_size):
        """ Setup batch size to cursor for this queryset.
        """
        if batch_size is not None:
            return self.copy(batch_size=batch_size)
        else:
            qs = self.copy()
            qs._batch_size = None
            return qs

    def __iter__(self):
        raise NotImplementedError

    def __contains__(self, document):
        raise NotImplementedError

    def find_one(self, criteria=None, projection=None, *, exc=None):
        raise NotImplementedError

    def update(self, update, *, multi=True, upsert=False):
        raise NotImplementedError

    def find_and_modify(
            self, update=None, *, upsert=False,
            full_response=False, new=False,
            **kwargs):
        raise NotImplementedError

    def remove(self, *, multi=True):
        raise NotImplementedError

    def distinct(self, field):
        raise NotImplementedError

    def count(self):
        raise NotImplementedError

    def ids(self):
        raise NotImplementedError

    def join(self, *field_names):
        raise NotImplementedError

    def find_in(self, comparable, field='_id', *,
                not_found=NotFoundBehavior.SKIP):
        raise NotImplementedError


class QuerySet(BaseQuerySet):
    def __iter__(self):
        return self._from_mongo_list(self._cursor)

    def __len__(self):
        return self.count()

    def __contains__(self, document):
        return self.find_one(document.id) is not None

    def _get_one(self, index):
        return self._from_mongo_one(self._cursor[index])

    def _from_mongo_list(self, data):
        """ Generator for got documents from raw data list (cursor).
        """
        for d in data:
            yield self._from_mongo_one(d)

    def find_one(self, criteria=None, projection=None, *, exc=None):
        """ Find and return only one document.

        :param dict criteria: update queryset's criteria
        :param dict projection: update queryset's projection
        :param Exception exc: raise given exception if not found
        :return: :class:`yadm.documents.Document` or **None**

        .. code:: python

            qs({'field': {'$gt': 3}}, {'field': True})
        """
        if isinstance(criteria, ObjectId):
            criteria = {'_id': criteria}

        qs = self.find(criteria=criteria, projection=projection)
        data = self._collection.find_one(qs._criteria, qs._projection)

        if data is None:
            if exc is not None:
                raise exc(qs)
            else:
                return None

        return self._from_mongo_one(data, projection=qs._projection)

    def with_id(self, _id):
        """ Find document with id.

        This method is deprecated. Use find_one.

        :param _id: id of searching document
        :return: :class:`yadm.documents.Document` or **None**
        """
        import warnings
        warnings.warn("QuerySet.with_id is deprecated", DeprecationWarning)
        doc = self._document_class()
        doc._id = _id
        return self.find_one({'_id': doc._id})

    def update(self, update, *, multi=True, upsert=False):
        """ Update documents in queryset.

        :param dict update: update query
        :param bool multi: update all matched documents
            *(default True)*
        :param bool upsert: insert if not found
            *(default False)*
        :return: update result
        """
        raw_result = self._collection.update(
            self._criteria,
            update,
            multi=multi,
            upsert=upsert,
        )
        return UpdateResult(raw_result)

    def find_and_modify(
            self, update=None, *, upsert=False,
            full_response=False, new=False,
            **kwargs):
        """ Execute *$findAndModify* query.

        :param dict update: see second argument to update()
        :param bool upsert: insert if object doesn’t exist
            *(default False)*
        :param bool full_response: return the entire response
            object from the server *(default False)*
        :param new: return updated rather than original object
            *(default False)*
        :param kwargs: any other options the findAndModify
            command supports can be passed here
        :return: :class:`yadm.documents.Document` or **None**
        """
        result = self._collection.find_and_modify(
            query=self._criteria,
            update=update,
            upsert=upsert,
            sort=self._sort or [],
            full_response=full_response,
            new=new,
            **kwargs
        )
        if not full_response:
            return self._from_mongo_one(result)
        else:
            result['value'] = self._from_mongo_one(result['value'])
            return result

    def remove(self, *, multi=True):
        """ Remove documents in queryset.

        :param bool multi: if False, remove only first finded document
            *(default True)*
        """
        raw_result = self._collection.remove(self._criteria, multi=multi)
        return RemoveResult(raw_result)

    def distinct(self, field):
        """ Distinct query.

        :param str field: field for distinct
        :return: list with result data
        """
        return self._cursor.distinct(field)

    def count(self):
        """ Count documents in queryset.

        :return: **int**
        """
        return self._cursor.count()

    def ids(self):
        """ Return all objects ids from queryset.
        """
        for raw in self.copy(projection={'_id': True})._cursor:
            yield raw['_id']

    def bulk(self):
        """ Return map {id: object}.

        :return: **dict**
        """
        qs = self.copy()
        qs._sort = None
        return {obj.id: obj for obj in qs}

    def join(self, *field_names):
        """ Create `yadm.Join` object, join `field_names` and return it.

        :param str fields_names: fields for join
        :return: new :class:`yadm.join.Join`

        Next algorithm for join:
            1. Get all documents from queryset;
            2. Aggegate all ids from requested fields;
            3. Make *$in* queries for get joined documents;
            4. Bind joined documents to objects from first queryset;

        `Join` object is instance of `abc.Sequence`.
        """
        join = Join(self)

        if field_names:
            join.join(*field_names)

        return join

    def find_in(self, comparable, field='_id', *,
                not_found=NotFoundBehavior.SKIP):
        """ Build ordered $in-query.

        Creates a query of the form {field: {'$in': comparable}} and
        returns the generator of documents with the same order as an elements
        in the argument 'comparable'.

        :param list comparable: values for compare in a query
        :param str field: field name of the document for comparison
        :param not_found: flag determines the behavior if the document
            with the specified value is not found
        :return: generator of docs

        not_found argument can take the following values:
            'none': If a document can not be found then a generator
                will return `None`.
            'skip': if a document can not be found then a generator
                will pass element.
            'error': if a document can not be found then a generator
                raise :class:`yadm.queryset.DocNotFoundError` exception.
        """
        not_found = NotFoundBehavior(not_found)
        hash_docs = {}

        for doc in self.find({field: {'$in': comparable}}):
            key = getattr(doc, field)
            if key not in hash_docs:
                hash_docs[key] = doc

        for cmp_item in comparable:
            value = hash_docs.get(cmp_item)

            if not_found is NotFoundBehavior.NONE:
                yield value

            elif not_found is NotFoundBehavior.SKIP:
                if value is not None:
                    yield value

            elif not_found is NotFoundBehavior.ERROR:
                if value is not None:
                    yield value
                else:
                    raise NotFoundError("Could not find a document with"
                                        " the field '{}' equal '{}'"
                                        "".format(field, cmp_item))



import pytest

from yadm import fields
from yadm.documents import Document


class Doc(Document):
    __collection__ = 'testdoc'
    e = fields.SimpleEmbeddedDocumentField({
        'i': fields.IntegerField(),
        's': fields.StringField(default='default value'),
    })


def test_init():
    doc = Doc()

    assert not hasattr(doc.e, 'i')
    assert doc.e.s == 'default value'


def test_init__type_error():
    with pytest.raises(TypeError):
        fields.SimpleEmbeddedDocumentField(object())


def test_init__empty():
    with pytest.raises(ValueError):
        fields.SimpleEmbeddedDocumentField({})


def test_save(db):
    doc = Doc()
    doc.e.i = 13
    db.save(doc)

    raw = db.db['testdoc'].find_one()

    assert raw['e']['i'] == 13
    assert raw['e']['s'] == 'default value'


def test_load(db):
    db.db['testdoc'].insert_one({'e': {'i': 26}})

    doc = db(Doc).find_one()

    assert doc.e.i == 26
    assert not hasattr(doc.e, 's')



import pytest

from bson import ObjectId

from yadm.documents import Document
from yadm import fields
from yadm.fields.references_list import (
    ReferencesListField,
    ReferencesList,
    NotResolved,
    AlreadyResolved,
)


class RDoc(Document):
    __collection__ = 'testrdocs'
    i = fields.IntegerField()


class Doc(Document):
    __collection__ = 'testdocs'
    ref = ReferencesListField(RDoc)


@pytest.fixture(scope='function')
def docs(db):
    ref_one = RDoc()
    ref_one.i = 13
    db.insert(ref_one)

    ref_two = RDoc()
    ref_two.i = 42
    db.insert(ref_two)

    ref_three = RDoc()
    ref_three.i = 666
    db.insert(ref_three)

    doc_one = Doc()
    doc_one.ref.append(ref_one)
    doc_one.ref.append(ref_two)
    db.insert(doc_one)

    doc_two = Doc()
    doc_two.ref.append(ref_two)
    db.insert(doc_two)

    doc_three = Doc()
    doc_three.ref.append(ref_two)
    doc_three.ref.append(ref_three)
    db.insert(doc_three)

    return [doc_one, doc_two, doc_three]


def test_save_empty(db):
    doc = Doc()
    db.insert(doc)

    data = db.db.testdocs.find_one()
    assert data['ref'] == []


def test_save_new(db):
    doc = Doc()
    assert doc.ref is doc.ref

    ref_one = RDoc()
    ref_one.i = 13
    db.insert(ref_one)

    doc.ref.append(ref_one)

    ref_two = RDoc()
    ref_two.i = 42
    db.insert(ref_two)
    doc.ref.append(ref_two)

    db.insert(doc)

    data = db.db.testdocs.find_one()

    assert len(data['ref']) == 2
    assert isinstance(data['ref'][0], ObjectId)
    assert isinstance(data['ref'][1], ObjectId)
    assert data['ref'][0] == ref_one.id
    assert data['ref'][1] == ref_two.id


def test_resolve(db, docs):
    doc = db.get_document(Doc, docs[0].id)
    assert isinstance(doc.ref, ReferencesList)
    assert len(doc.ref) == 2
    assert len(doc.ref.ids) == 2
    assert len(doc.ref._documents) == 0
    assert not doc.ref.resolved
    assert not doc.ref.changed

    doc.ref.resolve()
    assert doc.ref.resolved
    assert not doc.ref.changed
    assert len(doc.ref) == 2
    assert len(doc.ref.ids) == 2
    assert len(doc.ref._documents) == 2

    for rdoc in doc.ref:
        assert isinstance(rdoc, RDoc)


def test_already_resolved(db, docs):
    doc = db.get_document(Doc, docs[0].id)
    doc.ref.resolve()

    with pytest.raises(AlreadyResolved):
        doc.ref.resolve()


def test_not_resolved(db, docs):
    doc = db.get_document(Doc, docs[0].id)
    rdoc = RDoc(_id=ObjectId(), i=10)

    with pytest.raises(NotResolved):
        doc.ref[0]

    with pytest.raises(NotResolved):
        doc.ref[0] = rdoc

    with pytest.raises(NotResolved):
        doc.ref.insert(0, rdoc)

    with pytest.raises(NotResolved):
        doc.ref.append(rdoc)

    with pytest.raises(NotResolved):
        doc.ref.pop(0)

    with pytest.raises(NotResolved):
        for doc in doc.ref:
            pass

    assert not doc.ref.resolved
    assert not doc.ref.changed


def test_repr(db, docs):
    doc = db.get_document(Doc, docs[1].id)
    assert '[{!s}]'.format(docs[1].ref[0].id) in repr(doc.ref)
    doc.ref.resolve()
    assert '[{!r}]'.format(docs[1].ref[0]) in repr(doc.ref)


def test_setattr__reflist(db, docs):
    doc = Doc()
    doc.ref = docs[0].ref

    assert doc.ref == docs[0].ref


def test_setattr__list(db, docs):
    doc = Doc()
    doc.ref = list(docs[0].ref)

    assert doc.ref == docs[0].ref


def test_getitem():
    rdoc = RDoc(_id=ObjectId(), i=10)
    obj = ReferencesList(RDoc, [])
    obj.append(rdoc)

    assert obj[0] == rdoc


def test_setitem():
    rdoc_1 = RDoc(_id=ObjectId(), i=10)
    obj = ReferencesList(RDoc, [])
    obj.append(rdoc_1)

    rdoc_2 = RDoc(_id=ObjectId(), i=10)
    obj[0] = rdoc_2

    assert obj.ids == [rdoc_2.id]
    assert obj._documents == [rdoc_2]


def test_delitem():
    rdoc = RDoc(_id=ObjectId(), i=10)
    obj = ReferencesList(RDoc, [])
    obj.append(rdoc)
    del obj[0]

    assert len(obj) == 0
    assert not obj


def test_append():
    rdoc = RDoc(_id=ObjectId(), i=10)
    obj = ReferencesList(RDoc, [])
    obj.append(rdoc)

    assert obj.ids == [rdoc.id]
    assert obj._documents == [rdoc]


def test_insert():
    rdoc_1 = RDoc(_id=ObjectId(), i=10)
    rdoc_2 = RDoc(_id=ObjectId(), i=12)
    rdoc_3 = RDoc(_id=ObjectId(), i=13)

    obj = ReferencesList(RDoc, [])
    obj.insert(0, rdoc_1)
    obj.insert(0, rdoc_2)
    obj.insert(1, rdoc_3)

    assert obj.ids == [rdoc_2.id, rdoc_3.id, rdoc_1.id]
    assert obj._documents == [rdoc_2, rdoc_3, rdoc_1]


def test_pop():
    rdoc_1 = RDoc(_id=ObjectId(), i=10)
    rdoc_2 = RDoc(_id=ObjectId(), i=12)

    obj = ReferencesList(RDoc, [])
    obj.append(rdoc_1)
    obj.append(rdoc_2)

    res = obj.pop(1)
    assert res == rdoc_2
    assert obj.ids == [rdoc_1.id]
    assert obj._documents == [rdoc_1]



from yadm.cache import StackCache


def test_stack_getsetcontains():
    cache = StackCache(2)
    assert len(cache) == 0

    cache['a'] = 1
    cache['b'] = 2

    assert len(cache) == 2
    assert cache == {'a': 1, 'b': 2}
    assert cache['a'] == 1
    assert 'a' in cache
    assert 'c' not in cache


def test_stack_owerflow():
    cache = StackCache(2)
    assert len(cache) == 0

    cache['a'] = 1
    cache['b'] = 2
    cache['c'] = 3

    assert len(cache) == 2
    assert cache == {'b': 2, 'c': 3}
    assert 'a' not in cache
    assert 'c' in cache



from collections.abc import Sequence

from pymongo.errors import BulkWriteError

from yadm.common import BaseResult, build_update_query
from yadm.serialize import to_mongo, from_mongo


class Bulk:
    """ Bulk object.

    :param Database db: Database instance
    :param MetaDocument document_class: document class for collection
    :param bool ordered: create ordered bulk (default `False`)
    :param bool raise_on_errors: raise BulkWriteError exception
        if write errors (default `True`)

    Context manager example:

        with db.bulk(Doc, ordered=True) as bulk:
            bulk.insert(doc_1)
            bulk.insert(doc_2)
            bulk.update_one(doc_3, inc={'incr_key': 1})
            bulk.find({'key': 'value'}).update(set={'key': 'new_value'})
            bulk.find({'key': 'new_value'}).remove()
    """
    _result = None
    _error = False

    def __init__(self, db, document_class,
                 ordered, raise_on_errors,
                 collection_params):
        self._db = db
        self._document_class = document_class
        self._ordered = ordered
        self._raise_on_errors = raise_on_errors

        self._collection = db._get_collection(self._document_class,
                                              params=collection_params)

        if ordered:
            self._bulk_mongo = self._collection.initialize_ordered_bulk_op()
        else:
            self._bulk_mongo = self._collection.initialize_unordered_bulk_op()

    def __repr__(self):
        return "<{}.{}('{}') at {}>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            self._collection.name,
            hex(id(self)),
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.execute()

    @property
    def result(self):
        """ A BulkResult instance or rise RuntimeError if not executed.
        """
        if self._result is not None:
            return self._result
        else:
            raise RuntimeError("The bulk was not executed.")

    @property
    def error(self):
        """ True for executed errors.
        """
        return self._error

    def execute(self):
        """ Execute the bulk query.

        :return: :py:class:`BulkResult` instance
        """
        try:
            raw_data = self._bulk_mongo.execute()
        except BulkWriteError as exc:
            self._error = True
            raw_data = exc.details

        self._result = BulkResult(self, raw_data)

        if self._error and self._raise_on_errors:
            raise BulkWriteError(raw_data)

        return self._result

    def insert(self, document):
        """ Add insert document to bulk.

        :param Document document: document for insert

        .. warning::
            This unlike :py:class:`Database.insert <yadm.database.Database.insert>`!
            Currently, it is not bind objects to database and set id.
        """
        if not isinstance(document, self._document_class):
            raise TypeError("Bulk.insert() argument must be a {}, not '{}'"
                            "".format(self._document_class, document.__class__))

        self._bulk_mongo.insert(to_mongo(document))

    def find(self, query):
        """ Start "find" query in bulk.

        :param dict query:
        :return: BulkQuery instance
        """
        return BulkQuery(self._bulk_mongo, query)

    def update_one(self, document, *,
                   set=None, unset=None, inc=None,
                   push=None, pull=None):
        """ Add update one document to bulk.
        """
        if not isinstance(document, self._document_class):
            raise TypeError("Bulk.update_one() first argument must be a {}, not '{}'"
                            "".format(self._document_class, document.__class__))

        self.find({'_id': document.id}).update(set=set, unset=unset, inc=inc,
                                               push=push, pull=pull)


class BulkQuery:
    def __init__(self, bulk_mongo, query, *, upsert=False):
        self._bulk_mongo = bulk_mongo
        self._query = query
        self._upsert = upsert

    def __repr__(self):
        if not self._upsert:
            return "{s.__class__.__name__}({s._query!r})".format(s=self)
        else:
            return "{s.__class__.__name__}({s._query!r} upsertable)".format(s=self)

    def upsert(self):
        return self.__class__(self._bulk_mongo, self._query, upsert=True)

    def update(self, *,
               set=None, unset=None, inc=None,
               push=None, pull=None):
        """ Updale documents finded in query.
        """
        update_query = build_update_query(set=set, unset=unset, inc=inc,
                                          push=push, pull=pull)

        if self._upsert:
            self._bulk_mongo.find(self._query).upsert().update(update_query)
        else:
            self._bulk_mongo.find(self._query).update(update_query)

    def remove(self):
        self._bulk_mongo.find(self._query).remove()


class BulkResult(BaseResult):
    """ Object who provide result of `Bulk.execute()`.
    """
    def __init__(self, bulk, raw):
        super().__init__(raw)
        self._bulk = bulk

    def __bool__(self):
        return not self._bulk.error

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self._raw)

    @property
    def n_inserted(self):
        """ Provide `nInserted` from raw result.
        """
        return self['nInserted']

    @property
    def n_upserted(self):
        """ Provide `nUpserted` from raw result.
        """
        return self['nUpserted']

    @property
    def n_modified(self):
        """ Provide `nModified` from raw result.
        """
        return self['nModified']

    @property
    def n_removed(self):
        """ Provide `nRemoved` from raw result.
        """
        return self['nRemoved']

    @property
    def write_errors(self):
        """ Provide `writeErrors` from raw result.
        """
        return _BulkResultWriteErrors(self._bulk, self._raw['writeErrors'])

    @property
    def write_concern_errors(self):
        # TODO
        return self._raw['writeConcernErrors']

    @property
    def upserted(self):
        # TODO
        return self._raw['upserted']


class _BulkResultWriteErrors(Sequence):
    def __init__(self, bulk, raw_data):
        self._bulk = bulk
        self._raw = raw_data

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            raise TypeError("Slicing not supported")
        else:
            return _BulkResultWriteErrorsItem(self._bulk, self._raw[idx])

    def __len__(self):
        return len(self._raw)


class _BulkResultWriteErrorsItem(BaseResult):
    _document_cache = None

    def __init__(self, bulk, raw):
        super().__init__(raw)
        self._bulk = bulk

    @property
    def index(self):
        return self['index']

    @property
    def code(self):
        return self['code']

    @property
    def errmsg(self):
        return self['errmsg']

    @property
    def op(self):
        return self['op']

    @property
    def document(self):
        if self._document_cache is None:
            document_class = self._bulk._document_class
            self._document_cache = from_mongo(document_class, self.op)

        return self._document_cache



from collections.abc import Mapping


class Result(Mapping):
    def __init__(self, raw):
        self._raw = raw

    @property
    def raw(self):
        return self._raw

    def __getitem__(self, key):
        return self.raw[key]

    def __iter__(self):
        return iter(self.raw)

    def __len__(self):
        return len(self.raw)


class UpdateResult(Result):
    def __int__(self):
        return self.modified

    def __bool__(self):
        return self.ok

    @property
    def ok(self):
        return bool(self.raw['ok'])

    @property
    def matched(self):
        return self.raw['n']

    @property
    def upserted(self):
        return self.raw.get('nUpserted', 0)

    @property
    def modified(self):
        return self.raw.get('nModified', 0)


class RemoveResult(Result):
    def __int__(self):
        return self.removed

    def __bool__(self):
        return self.ok

    @property
    def ok(self):
        return bool(self.raw['ok'])

    @property
    def removed(self):
        return self.raw['n']



import pytest

from yadm import (
    Document,
    EmbeddedDocument,
)
from yadm.common import EnclosedDocDescriptor
from yadm.fields import (
    Field,
    BooleanField,
    IntegerField,
    FloatField,
    StringField,
    EmbeddedDocumentField,
    ReferenceField,
)


class Foo(Document):
    f = BooleanField(True)


class Bar(EmbeddedDocument):
    f = IntegerField(42)


class Baz(Document):
    __collection__ = 'baz'
    f = FloatField(3.14)


class Qux(EmbeddedDocument):
    f = StringField('quux')


class Corge(Document):
    grault = EmbeddedDocumentField(Foo)
    garply = ReferenceField(Bar)
    waldo = EmbeddedDocumentField(Baz)
    fred = ReferenceField(Qux)
    plugh = EmbeddedDocumentField('self')
    xyzzy = ReferenceField('self')


class Thud(Document):
    rf = ReferenceField('dumb.path')


class Thud2(EmbeddedDocument):
    ef = EmbeddedDocumentField('another.dumb.path')


class FooDocField(Field):
    reference_document_class = EnclosedDocDescriptor('reference')
    embedded_document_class = EnclosedDocDescriptor('embedded')


@pytest.fixture(scope='function')
def fs(request):
    class FieldsSet:
        R = ReferenceField(Foo)
        E = EmbeddedDocumentField(Bar)
        CR = ReferenceField('tests.test_common.Baz')
        CE = EmbeddedDocumentField('tests.test_common.Qux')

    return FieldsSet


def assert_reference(classinfo, expected_value):
    assert classinfo.reference_document_class == expected_value
    assert getattr(classinfo, '_reference_document_class') == expected_value


def assert_embedded(classinfo, expected_value):
    assert classinfo.embedded_document_class == expected_value
    assert getattr(classinfo, '_embedded_document_class') == expected_value


def test_reference(fs):
    assert_reference(fs.R, Foo)


def test_embedded(fs):
    assert_embedded(fs.E, Bar)


def test_resolving(fs):
    assert_reference(fs.CR, Baz)
    assert_embedded(fs.CE, Qux)

    assert_embedded(Corge.grault, Foo)
    assert_reference(Corge.garply, Bar)
    assert_embedded(Corge.waldo, Baz)
    assert_reference(Corge.fred, Qux)
    assert_embedded(Corge.plugh, Corge)
    assert_reference(Corge.xyzzy, Corge)

    with pytest.raises(ImportError):
        Thud.rf.reference_document_class

    with pytest.raises(ImportError):
        Thud2.ef.embedded_document_class


def test_lowlevel(fs):
    EnclosedDocDescriptor('reference')
    EnclosedDocDescriptor('embedded')

    for cls_type in ('waldo', 'fred'):
        with pytest.raises(ValueError):
            EnclosedDocDescriptor(cls_type)

    # class binding
    assert isinstance(FooDocField.reference_document_class,
                      EnclosedDocDescriptor)
    assert isinstance(FooDocField.embedded_document_class,
                      EnclosedDocDescriptor)

    # getattr --> None
    doc_field = FooDocField()
    assert doc_field.reference_document_class is None
    assert doc_field.embedded_document_class is None

    doc_field.reference_document_class = Foo
    doc_field.embedded_document_class = Bar

    assert_reference(doc_field, Foo)
    assert_embedded(doc_field, Bar)

    # destructor
    del doc_field.reference_document_class
    assert getattr(doc_field, '_reference_document_class', None) is None
    del doc_field.embedded_document_class
    assert getattr(doc_field, '_embedded_document_class', None) is None


def test_redefine_notloaded(db):
    baz = Baz(f=13.0)
    db.save(baz)
    baz = db(Baz).fields('_id').find_one({'f': 13.0})
    baz.f = 666.0
    db.save(baz)



""" YADM with faker integration.
"""
from collections import Counter
from types import GeneratorType

import pymongo
from faker import Faker

from yadm.documents import BaseDocument, Document, EmbeddedDocument
from yadm.markers import AttributeNotSet


DEFAULT_DEPTH = 4  # <=450


def create_fake(__document_class__,
                __db__=None,
                __parent__=None,
                __name__=None,
                __faker__=None,
                __depth__=DEFAULT_DEPTH,
                __write_concern__=pymongo.WriteConcern(w='majority'),
                **values):
    """ Create document with fake data.

    :param yadm.documents.BaseDocument __document_class__: document class
        for new instance
    :param yadm.database.Database __db__: database instance
        if specified, document and all references will be saved to database
    :param yadm.documents.BaseDocument __parent__: parent document
    :param str __name__: name of parent field
    :param Faker __faker__: faker instance, create if not specified
    :param int __depth__: maximum recursion depth,
        not recomendated use greater than 450
        (default 4)
    :return yadm.documents.BaseDocument: __document_class__ instance with fake data
    """
    if not issubclass(__document_class__, BaseDocument):
        raise TypeError("only BaseDocument subclasses is allowed")

    create_fake.counter[__document_class__] += 1

    if __depth__ < 0:
        return AttributeNotSet

    if __faker__ is None:
        __faker__ = Faker()

    document = __document_class__()

    if isinstance(document, Document):
        document.__db__ = __db__
    elif isinstance(document, EmbeddedDocument):
        document.__parent__ = __parent__
        document.__name__ = __name__

    doc_fake_proc = document.__fake__(values, __faker__, __depth__ - 1)

    # extend values from __fake__ method
    if isinstance(doc_fake_proc, GeneratorType):
        values = next(doc_fake_proc)
    elif isinstance(doc_fake_proc, dict):
        values = doc_fake_proc

    # first: set values
    for name, fake in values.items():
        if fake is not AttributeNotSet:
            setattr(document, name, fake)

    # second: field faker
    for name, field in __document_class__.__fields__.items():
        if name not in values and not hasattr(document, '__fake__{}__'.format(name)):
            fake = field.get_fake(document, __faker__, __depth__ - 1)
            if fake is not AttributeNotSet:
                setattr(document, name, fake)

    # third: __fake__{name}__ methods
    for name, field in __document_class__.__fields__.items():
        if name not in values and hasattr(document, '__fake__{}__'.format(name)):
            attr = getattr(document, '__fake__{}__'.format(name))
            fake = attr(__faker__, __depth__ - 1)

            if fake is not AttributeNotSet:
                setattr(document, name, fake)

    if isinstance(doc_fake_proc, GeneratorType):
        # pre save processor
        try:
            next(doc_fake_proc)
        except StopIteration:
            doc_fake_proc = None

    if __db__ is not None:
        __db__.insert(document, write_concern=__write_concern__)

        # post save processor
        if isinstance(doc_fake_proc, GeneratorType):
            try:
                next(doc_fake_proc)
            except StopIteration:
                pass

    return document


create_fake.counter = Counter()



from collections.abc import Mapping

from yadm.fields.base import Field


class UnmutableMap(Mapping):
    def __init__(self, data):
        self._data = data

    def __getitem__(self, item):
        return self._data[item]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return repr("{}({!r})".format(self.__class__.__name__, self._data))


class MongoMapField(Field):
    def __init__(self, smart_null=False):
        super().__init__(smart_null=smart_null)

    def get_fake(self, document, faker, deep):  # pragma: no cover
        return UnmutableMap({})

    def prepare_value(self, document, value):
        return UnmutableMap(value)

    def from_mongo(self, document, value):
        return UnmutableMap(value)

    def to_mongo(self, document, value):
        return dict(value)



from random import randint
from unittest.mock import Mock, MagicMock

import pytest

from yadm import Document
from yadm import fields
from yadm.aggregation import Aggregator, AgOperator


class Doc(Document):
    __collection__ = 'docs'
    i = fields.IntegerField()


@pytest.fixture(scope='function')
def docs(db):
    with db.bulk(Doc) as bulk:
        docs = []
        for n in range(randint(10, 20)):
            doc = Doc(i=randint(-666, 666))
            bulk.insert(doc)
            docs.append(doc)

    return docs


@pytest.fixture(scope='module')
def fake_aggregator():
    Doc = MagicMock(name='Doc')  # noqa
    Doc.__collection__ = 'col'
    return Aggregator(Mock(), Doc)


def test_pipeline(db, docs):
    agg = db.aggregate(Doc).match(i={'$gt': 0}).group({'_id': '1', 's': {'$sum': '$i'}})
    assert agg._pipeline == [
        {'$match': {'i': {'$gt': 0}}},
        {'$group': {'_id': '1', 's': {'$sum': '$i'}}},
    ]


def test_first(db, docs):
    agg = db.aggregate(Doc).group({
        '_id': '1',
        's': {'$sum': '$i'},
        'c': {'$sum': 1},
    })
    res = agg[0]

    assert set(res) == {'_id', 'c', 's'}
    assert res['c'] == len(docs)
    assert res['s'] == sum(d.i for d in docs)


def test_index(db, docs):
    agg = db.aggregate(Doc).sort(i=1)
    result = []
    for idx, _ in enumerate(agg):
        result.append(agg[idx]['i'])

    assert result == sorted([d.i for d in docs])


def test_indexerror(db, docs):
    agg = db.aggregate(Doc)

    with pytest.raises(IndexError):
        agg[100500]


def test_negativeindex(fake_aggregator):
    with pytest.raises(NotImplementedError):
        fake_aggregator[-1]


def test_slice(db, docs):
    numbers = sorted([d.i for d in docs])
    agg = db.aggregate(Doc).sort(i=1)
    assert [d['i'] for d in agg[1:3]] == numbers[1:3]


def test_slicestep(fake_aggregator):
    with pytest.raises(NotImplementedError):
        fake_aggregator[1:3:2]


def test_indextypeerror(fake_aggregator):
    with pytest.raises(TypeError):
        fake_aggregator['1']


def test_for(db, docs):
    agg = db.aggregate(Doc).match(i={'$gt': 0}).project(n='$i')
    count = 0

    for item in agg:
        assert item['n'] > 0
        count += 1

    assert count == len([d.i for d in docs if d.i > 0])


def test_repr(fake_aggregator):
    assert 'Aggregator' in repr(fake_aggregator)
    assert 'col' in repr(fake_aggregator)

    agop = AgOperator(fake_aggregator, 'operator')
    assert 'AgOperator' in repr(agop)
    assert 'operator' in repr(agop)
    assert 'col' in repr(agop)


def test_empty(fake_aggregator):
    with pytest.raises(ValueError):
        fake_aggregator.match()


def test_error(fake_aggregator):
    with pytest.raises(ValueError):
        fake_aggregator.match({'a': 1}, b=2)



"""
Mongo Aggregation Framework helper.

.. code-block:: python

    cur = db.aggregate(Doc).match({'i': {'$gt': 13}}).project(a='$i').limit(8)
"""

import pymongo

PYMONGO_VERSION = pymongo.version_tuple


class BaseAggregator:
    def __init__(self, db, document_class, *,
                 pipeline=None, collection_params=None):
        self._db = db
        self._document_class = document_class
        self._pipeline = [] if pipeline is None else pipeline
        self._collection_params = collection_params

    def __repr__(self):
        return ("{s.__class__.__name__}("
                "{s._document_class.__collection__} {s._pipeline})"
                "".format(s=self))

    def __getattr__(self, op):
        return AgOperator(self, op)

    @property
    def _cursor(self):  # noqa
        """ pymongo aggregate cursor.
        """
        collection = self._db._get_collection(self._document_class)
        cur = collection.aggregate(self._pipeline)
        if PYMONGO_VERSION < (3, 0):
            return cur['result']
        else:
            return cur


class Aggregator(BaseAggregator):
    def __iter__(self):
        return iter(self._cursor)

    def __getitem__(self, index):
        if isinstance(index, int):
            if index > 0:
                try:
                    return self.skip(index).limit(1)[0]
                except IndexError:
                    raise IndexError("index out of range: {}".format(index))

            elif index == 0:
                try:
                    return next(iter(self))
                except StopIteration:
                    raise IndexError("index out of range: 0")

            else:
                raise NotImplementedError("negative indexed are not implemented")

        elif isinstance(index, slice):
            if index.step is None or index.step == 1:
                start = index.start
                stop = index.stop
                return self.skip(start).limit(stop - start)
            else:
                raise NotImplementedError("stepped slice is not implemented")

        else:
            raise TypeError("{s.__class__.__name__}"
                            " indices must be integers,"
                            " not {t.__name__}"
                            "".format(s=self, t=type(index)))


class AgOperator:
    def __init__(self, aggregate, op):
        self._aggregate = aggregate
        self._op = op if op.startswith('$') else '${}'.format(op)

    def __call__(self, _value=None, **kwargs):
        if _value is not None and kwargs:
            raise ValueError()

        value = kwargs if kwargs else _value
        if not value:
            raise ValueError("empty value")

        pipeline = self._aggregate._pipeline.copy()
        pipeline.append({self._op: value})
        return self._aggregate.__class__(self._aggregate._db,
                                         self._aggregate._document_class,
                                         pipeline=pipeline)

    def __repr__(self):
        return ("{s.__class__.__name__}"
                "({a._document_class.__collection__} {a._pipeline} {s._op})"
                "".format(s=self, a=self._aggregate))



import pytest

from bson import ObjectId
import pymongo

from yadm.documents import Document
from yadm.queryset import QuerySet
from yadm.serialize import from_mongo
from yadm import fields


class Doc(Document):
    __collection__ = 'testdocs'
    b = fields.BooleanField()
    i = fields.IntegerField()
    l = fields.ListField(fields.IntegerField())


def test_init(db):
    assert isinstance(db.client, pymongo.mongo_client.MongoClient)
    assert isinstance(db.db, pymongo.database.Database)
    assert db.db.name == 'test'
    assert db.name == 'test'


def test_get_collection(db):
    collection = db._get_collection(Doc)

    assert isinstance(collection, pymongo.collection.Collection)
    assert collection.name == 'testdocs'


def test_get_queryset(db):
    queryset = db.get_queryset(Doc)

    assert isinstance(queryset, QuerySet)
    assert queryset._db is db
    assert queryset._document_class is Doc


def test_get_document(db):
    col = db.db.testdocs
    ids = [col.insert_one({'i': i}).inserted_id for i in range(10)]

    _id = ids[5]
    doc = db.get_document(Doc, _id)

    assert doc is not None
    assert doc._id == _id
    assert doc.i == 5
    assert doc.__db__ is db


def test_get_document_w_projection(db):
    col = db.db.testdocs
    ids = [col.insert_one({'i': i, 'b': bool(i % 2)}).inserted_id
           for i in range(10)]

    _id = ids[5]
    doc = db.get_document(Doc, _id, projection={'b': False})

    assert doc is not None
    assert doc._id == _id
    assert doc.i == 5
    assert doc.__db__ is db

    with pytest.raises(fields.base.NotLoadedError):
        doc.b


def test_get_document__not_found(db):
    col = db.db.testdocs
    [col.insert_one({'i': i}).inserted_id for i in range(10)]

    doc = db.get_document(Doc, ObjectId())

    assert doc is None


def test_get_document__not_found_exc(db):
    col = db.db.testdocs
    [col.insert_one({'i': i}).inserted_id for i in range(10)]

    class Exc(Exception):
        pass

    with pytest.raises(Exc):
        db.get_document(Doc, ObjectId(), exc=Exc)


def test_insert(db):
    doc = Doc()
    doc.i = 13

    db.insert(doc)

    assert db.db.testdocs.find().count() == 1
    assert db.db.testdocs.find()[0]['i'] == 13
    assert not doc.__changed__
    assert doc.__db__ is db


def test_save_new(db):
    doc = Doc()
    doc.i = 13

    db.save(doc)

    assert db.db.testdocs.find().count() == 1
    assert db.db.testdocs.find()[0]['i'] == 13
    assert not doc.__changed__
    assert doc.__db__ is db


def test_save(db):
    col = db.db.testdocs
    col.insert({'i': 13})

    doc = from_mongo(Doc, col.find_one({'i': 13}))
    doc.i = 26

    col.update({'_id': doc.id}, {'$set': {'b': True}})

    db.save(doc)

    assert doc.i == 26
    assert doc.b is True
    assert db.db.testdocs.find().count() == 1
    assert db.db.testdocs.find()[0]['i'] == 26
    assert db.db.testdocs.find()[0]['b'] is True
    assert not doc.__changed__
    assert doc.__db__ is db


def test_save_full(db):
    col = db.db.testdocs
    col.insert({'i': 13})

    doc = from_mongo(Doc, col.find_one({'i': 13}))
    doc.i = 26

    col.update({'_id': doc.id}, {'$set': {'b': True}})

    db.save(doc, full=True)

    assert doc.i == 26
    assert not hasattr(doc, 'b')
    assert db.db.testdocs.find().count() == 1
    assert db.db.testdocs.find()[0]['i'] == 26
    assert 'b' not in db.db.testdocs.find()[0]
    assert not doc.__changed__
    assert doc.__db__ is db


@pytest.fixture(scope='function')
def doc(db):
    return db.insert(
        Doc(
            b=True,
            i=13,
            l=[1, 2, 3],
        ),
    )


@pytest.mark.parametrize('kwargs, result', [
    (
        {'set': {'i': 88}},
        {'b': True, 'i': 88, 'l': [1, 2, 3]},
    ),
    (
        {'unset': {'i': True}},
        {'b': True, 'l': [1, 2, 3]},
    ),
    (
        {'unset': ['b']},
        {'i': 13, 'l': [1, 2, 3]},
    ),
    (
        {'push': {'l': 33}},
        {'b': True, 'i': 13, 'l': [1, 2, 3, 33]},
    ),
    (
        {'pull': {'l': 2}},
        {'b': True, 'i': 13, 'l': [1, 3]},
    ),
    (
        {'inc': {'i': 653}},
        {'b': True, 'i': 666, 'l': [1, 2, 3]},
    ),
])
def test_update_one(db, doc, kwargs, result):
    db.update_one(doc, **kwargs)
    raw_doc = db.db['testdocs'].find_one(doc.id)

    assert raw_doc
    del raw_doc['_id']
    assert raw_doc == result


def test_remove(db):
    col = db.db.testdocs
    col.insert({'i': 13})

    doc = from_mongo(Doc, col.find_one({'i': 13}))

    assert col.count() == 1
    db.remove(doc)
    assert col.count() == 0


def test_reload(db):
    doc = Doc()
    doc.i = 1
    db.insert(doc)

    db.db.testdocs.update({'_id': doc.id}, {'i': 2})

    assert doc.i == 1
    new = db.reload(doc)
    assert doc is new
    assert doc.i == 2


def test_reload_new_instance(db):
    doc = Doc()
    doc.i = 1
    db.insert(doc)

    db.db.testdocs.update({'_id': doc.id}, {'i': 2})

    assert doc.i == 1
    new = db.reload(doc, new_instance=True)
    assert doc is not new
    assert doc.i == 1
    assert new.i == 2


def test_w_projection(db):
    doc = Doc()
    doc.i = 1
    doc.b = True
    db.insert(doc)

    db.db.testdocs.update({'_id': doc.id}, {'$set': {'i': 2}})

    assert doc.i == 1
    assert doc.b is True
    new = db.reload(doc, projection={'b': False})
    assert doc is new
    assert doc.i == 2

    with pytest.raises(fields.base.NotLoadedError):
        doc.b



from bson import ObjectId

from yadm import fields
from yadm.documents import Document


class RDoc(Document):
    __collection__ = 'testrdocs'
    i = fields.IntegerField()


class Doc(Document):
    __collection__ = 'testdocs'
    li = fields.ListField(fields.ReferenceField(RDoc))


def test_save(db):
    doc = Doc()

    ref_one = RDoc()
    ref_one.i = 13
    db.insert(ref_one)
    doc.li.append(ref_one)

    ref_two = RDoc()
    ref_two.i = 42
    db.insert(ref_two)
    doc.li.append(ref_two)

    db.insert(doc)

    data = db.db.testdocs.find_one()

    assert len(data['li']) == 2
    assert isinstance(data['li'][0], ObjectId)
    assert isinstance(data['li'][1], ObjectId)
    assert data['li'][0] == ref_one.id
    assert data['li'][1] == ref_two.id


def test_load(db):
    db.db.testdocs.insert(
        {'li': [
            db.db.testrdocs.insert({'i': 13}),
            db.db.testrdocs.insert({'i': 42}),
        ]}
    )

    doc = db.get_queryset(Doc).find_one()

    assert hasattr(doc, 'li')
    assert len(doc.li) == 2

    for item in doc.li:
        assert isinstance(item, RDoc)

    assert doc.li[0].i == 13
    assert doc.li[1].i == 42



from yadm import fields
from yadm.documents import Document


def test_point_init():
    point = fields.Point(1, 2)
    assert point.longitude == 1
    assert point.latitude == 2


def test_point_get():
    longitude, latitude = fields.Point(1, 2)
    assert longitude == 1
    assert latitude == 2


def test_point_to_mongo():
    point = fields.Point(1, 2)
    assert point.to_mongo() == {'type': 'Point', 'coordinates': [1, 2]}


def test_point_from_mongo():
    point = fields.Point.from_mongo({'type': 'Point', 'coordinates': [1, 2]})
    assert isinstance(point, fields.Point)
    assert point.longitude == 1
    assert point.latitude == 2


def test_multipoint_init():
    mpoint = fields.MultiPoint([fields.Point(1, 2), fields.Point(3, 4)])

    assert len(mpoint._points) == 2
    assert len(mpoint) == 2
    assert isinstance(mpoint[0], fields.Point)

    assert mpoint[0].longitude == 1
    assert mpoint[0].latitude == 2
    assert mpoint[1].longitude == 3
    assert mpoint[1].latitude == 4


def test_multipoint_to_mongo():
    mpoint = fields.MultiPoint([fields.Point(1, 2), fields.Point(3, 4)])
    assert mpoint.to_mongo() == \
        {'type': 'MultiPoint', 'coordinates': [[1, 2], [3, 4]]}


def test_multipoint_from_mongo():
    mpoint = fields.MultiPoint.from_mongo(
        {'type': 'MultiPoint', 'coordinates': [[1, 2], [3, 4]]}
    )
    assert isinstance(mpoint, fields.MultiPoint)
    assert len(mpoint) == 2
    assert isinstance(mpoint[0], fields.Point)

    assert mpoint[0].longitude == 1
    assert mpoint[0].latitude == 2
    assert mpoint[1].longitude == 3
    assert mpoint[1].latitude == 4


class Doc(Document):
    __collection__ = 'testdoc'
    point = fields.PointField()


def test_doc_insert(db):
    doc = Doc()
    doc.point = fields.Point(1, 2)
    db.insert(doc)

    raw = db.db.testdoc.find_one({'_id': doc.id})
    assert 'point' in raw
    assert raw['point'] == {'type': 'Point', 'coordinates': [1, 2]}


def test_doc_get(db):
    _id = db.db.testdoc.insert(
        {'point': {'type': 'Point', 'coordinates': [1, 2]}}
    )
    doc = db(Doc).find_one(_id)

    assert hasattr(doc, 'point')
    assert isinstance(doc.point, fields.Point)
    assert doc.point.longitude == 1
    assert doc.point.latitude == 2


def test_doc_set_object():
    doc = Doc()
    doc.point = fields.Point(1, 2)
    assert isinstance(doc.point, fields.Point)
    assert doc.point.longitude == 1
    assert doc.point.latitude == 2


def test_doc_set_geojson():
    doc = Doc()
    doc.point = {'type': 'Point', 'coordinates': [1, 2]}
    assert isinstance(doc.point, fields.Point)
    assert doc.point.longitude == 1
    assert doc.point.latitude == 2



"""
Work with references.

.. code-block:: python

    class RDoc(Document):
        i = fields.IntegerField()

    class Doc(Document):
        rdoc = fields.ReferenceField(RDoc)

    rdoc = RDoc()
    rdoc.i = 13
    db.insert(rdoc)

    doc = Doc()
    doc.rdoc = rdoc
    db.insert(doc)

    doc = db.get_queryset(Doc).find_one(doc.id)  # reload doc
    assert doc.rdoc.id == rdoc.id
    assert doc.rdoc.i == 13

Or with asyncio:

.. code-block:: python

    rdoc = await doc.rdoc
    assert rdoc.id == rdoc.id
    assert rdoc.i == 13
    assert doc.rdoc == rdoc.id

"""
from bson import ObjectId
from bson.errors import InvalidId

from yadm.common import EnclosedDocDescriptor
from yadm.markers import AttributeNotSet
from yadm.documents import Document, DocumentItemMixin
from yadm.fields.base import Field, pass_null
from yadm.testing import create_fake


class BrokenReference(Exception):
    """ Raise if referrenced document is not found.
    """


class NotBindingToDatabase(Exception):  # noqa
    """ Raise if set ObjectId insted referenced document
    to new document, who not binded to database.
    """


class ReferenceField(Field):
    """ Field for work with references.

    :param reference_document_class: class for refered documents
    """
    reference_document_class = EnclosedDocDescriptor('reference')

    def __init__(self, reference_document_class, **kwargs):
        super().__init__(**kwargs)
        self.reference_document_class = reference_document_class

    def get_default(self, document):
        if self.smart_null:
            return None
        else:
            return AttributeNotSet

    def get_fake(self, document, faker, depth):
        """ Try create referenced document.
        """
        res = create_fake(
            self.reference_document_class,
            __db__=document.__db__,
            __faker__=faker,
            __depth__=depth)

        if res is AttributeNotSet and self.smart_null:
            return None
        else:
            return res

    def copy(self):
        return self.__class__(self.reference_document_class,
                              smart_null=self.smart_null)

    @pass_null
    def prepare_value(self, document, value):
        if isinstance(value, Document):
            return value
        elif value is AttributeNotSet:
            return AttributeNotSet
        else:
            if isinstance(value, str):
                try:
                    value = ObjectId(value)
                except InvalidId:
                    pass

            return self.from_mongo(document, value)

    @pass_null
    def from_mongo(self, document, value):
        if document.__db__ is not None:
            rdc = self.reference_document_class

            if document.__qs__ is not None:
                cache = document.__qs__.cache
            else:
                cache = {}  # fake cache

            if (rdc, value) in cache:
                return cache[(rdc, value)]
            else:
                if document.__db__.aio:
                    cache[(rdc, value)] = ref = Reference(value, document, self)
                    return ref
                else:
                    qs = document.__db__.get_queryset(rdc, cache=cache)
                    doc = qs.find_one(value)
                    if doc is None:  # pragma: no cover
                        doc = qs.read_primary().find_one(value, exc=BrokenReference)

                    cache[(rdc, value)] = doc
                    return doc

        else:
            raise NotBindingToDatabase((document, self, value))

    @pass_null
    def to_mongo(self, document, value):
        return value.id


class Reference(ObjectId):
    """ Reference object.

    This is awaitable:

        doc = await doc.reference
    """
    document = None

    def __init__(self,
                 _id: ObjectId,
                 parent: DocumentItemMixin,
                 field: ReferenceField):
        super().__init__(_id)
        self.parent = parent
        self.field = field
        self.db = parent.__db__
        self.document_class = field.reference_document_class

    def __repr__(self):
        n = self.__class__.__name__
        collection = self.document_class.__collection__
        status = '+' if self.document is not None else '-'
        return "{}({}:{} {})".format(n, collection, str(self), status)

    def __await__(self):
        return self.get().__await__()

    async def get(self, force: bool=False):
        if self.document is None or force:
            self.document = await self.db(self.document_class).find_one(self)
            if self.document is None:  # pragma: no cover
                self.document = await self.db.get_document(self.document_class, self)

        return self.document



""" Field for money.

Use :class:`yadm.fields.money.Money`, as value for money

.. code block: python

    class DocClass(Document):
        money = MoneyField()

    doc = DocClass()
    doc.money = Money('3.14', 'USD')

    db.insert(doc)

This code save to MongoDB document:

.. code block: javascript

    {
        id: ObjectId('534272984c78591787e1a964'),
        money: {v: 314, c: 'USD'}
    }
"""
from decimal import Decimal, Context, ROUND_UP
from functools import wraps
import random

from yadm.fields.base import Field, DefaultMixin, pass_null
from .currency import DEFAULT_CURRENCY_STORAGE
from yadm.markers import AttributeNotSet


def _checker(method):
    """ Decorator for check type of value.
    """
    @wraps(method)
    def wrapper(self, value):
        if not isinstance(value, Money):
            return NotImplemented

        if self.currency != value.currency:
            raise ValueError("Money {!r} and {!r} must be with the"
                             " same currencies".format(str(self), str(value)))
        return method(self, value)

    return wrapper


def _special_comparison(method):
    """ Decorator for special compression Money with other values.
    """
    @wraps(method)
    def wrapper(self, value):
        if isinstance(value, int):
            # This need to compare Money with 0 only integer
            if value:
                return NotImplemented
            else:
                return method(self.value, value)

        elif isinstance(value, Money):
            if self.currency != value.currency:
                raise ValueError("Money {!r} and {!r} must be with"
                                 " the same currencies".format(self, value))
            else:
                return method(self.value, value.value)

        else:
            return NotImplemented

    return wrapper


class Money:
    _context = Context(rounding=ROUND_UP)

    def __init__(self, value, currency=None):
        if isinstance(value, Money):
            if currency is not None:
                raise TypeError("new Money from another Money not need"
                                " currency {!r} as parameter".format(currency))
            else:
                self._value = value.value
                self._currency = value.currency

        else:
            if currency is None:
                raise TypeError("Curency not set.")
            else:
                self._currency = currency = DEFAULT_CURRENCY_STORAGE[currency]

            if isinstance(value, Decimal):
                precision_decimal = Decimal('1.' + '0' * currency.precision)
                self._value = value.quantize(precision_decimal, self._context.rounding)

            else:
                value = Decimal(value, context=self._context)
                precision_decimal = Decimal('1.' + '0' * currency.precision)
                self._value = value.quantize(precision_decimal, self._context.rounding)

    @classmethod
    def from_cents(cls, cents: int, currency) -> 'Money':
        """ Return new Money object from cents.
        """
        _currency = DEFAULT_CURRENCY_STORAGE[currency]
        delimiter = Decimal(10 ** _currency.precision)
        return cls(cents / delimiter, _currency)

    @classmethod
    def from_string(cls, s):
        v, c = s.strip().split()

        if c.isdigit():
            c = int(c)
        else:
            ValueError(s)

        try:
            return cls(v, DEFAULT_CURRENCY_STORAGE[c])
        except KeyError as exc:
            raise ValueError from exc

    @property
    def value(self) -> Decimal:
        return self._value

    @property
    def currency(self):
        return self._currency

    @property
    def total_cents(self) -> int:
        """ Return total cents in this object.
        """
        precision = self.currency.precision
        precision_decimal = Decimal('1.' + '0' * precision)
        quantized = self.value.quantize(precision_decimal, self._context.rounding)
        return int(quantized * 10 ** precision)

    def __abs__(self):
        return self.__class__(abs(self._value), self.currency)

    @_checker
    def __add__(self, target):
        return self.__class__(self.value + target.value, self.currency)

    @_checker
    def __sub__(self, target):
        return self.__class__(self.value - target.value, self.currency)

    def __neg__(self):
        return Money(-self.value, self.currency)

    def __mul__(self, target):
        if isinstance(target, (int, Decimal)):
            return self.__class__(self.value * target, self.currency)
        else:
            return NotImplemented

    def __rmul__(self, target):
        return self.__mul__(target)

    def __truediv__(self, target):
        if isinstance(target, self.__class__):
            if self.currency == target.currency:
                return self.value / target.value
            else:
                raise ValueError("Money {!r} and {!r} must be with"
                                 " the same currencies".format(self, target))
        else:
            return self.__class__(self.value / target, self.currency)

    def __eq__(self, target):
        if target is 0:
            return self.value == target
        else:
            return (isinstance(target, self.__class__) and
                    self.value == target.value and
                    self.currency == target.currency)

    @_special_comparison
    def __gt__(self, other):
        return self > other

    @_special_comparison
    def __lt__(self, other):
        return self < other

    @_special_comparison
    def __ge__(self, other):
        return self > other or self == other

    @_special_comparison
    def __le__(self, other):
        return self < other or self == other

    def __bool__(self):
        return bool(self.value)

    def __hash__(self):
        return hash(str(self))

    def __str__(self):
        precision = self.currency.precision
        precision_decimal = Decimal('1.' + '0' * precision)
        quantized = self.value.quantize(precision_decimal, self._context.rounding)
        return '{} {}'.format(quantized, self._currency.string)

    def __repr__(self):
        return 'Money({!s})'.format(self)

    def to_mongo(self):
        # b/c
        return MoneyField.to_mongo(None, None, self)

    def is_positive(self):
        return self.value > 0

    def is_negative(self):
        return self.value < 0

    def is_zero(self):
        return self.value == 0


class MoneyField(DefaultMixin, Field):
    """ Field to storage money values.
    """
    def __init__(self, bcc=None, **kwargs):
        """ Field for save Money.

        bcc - str,int or Currency value to backward compatibility.
        """
        super().__init__(**kwargs)
        self._bcc = bcc

    def get_fake(self, document, faker, depth):
        value = faker.pydecimal(left_digits=5, right_digits=2, positive=True)
        currency = random.choice(list(DEFAULT_CURRENCY_STORAGE.values()))
        return Money(value, currency)

    @pass_null
    def prepare_value(self, document, value):
        if isinstance(value, Money):
            return value
        else:
            raise TypeError("Only money is allowed for asigment to MoneyField.")

    @pass_null
    def to_mongo(self, document, value):
        return [value.total_cents, value.currency.code]

    @pass_null
    def from_mongo(self, document, data):
        if data is AttributeNotSet:
            return AttributeNotSet

        if isinstance(data, int):
            if not self._bcc:
                raise ValueError("Backward compatibility currency not set in MoneyField.")
            else:
                currency_key = self._bcc
                value = data

        elif isinstance(data, list):
            value, currency_key = data

        elif isinstance(data, dict):  # b/c for version 1.3.0
            currency_key = data['c']
            value = data['v']

        else:
            raise TypeError("Incorrect type of value {!r} for Money".format(type(data)))

        currency = DEFAULT_CURRENCY_STORAGE[currency_key]
        new_value = Decimal(value) / (10 ** currency.precision)

        return Money(new_value, currency)



import pytest

from yadm import fields
from yadm.documents import Document


class Doc(Document):
    __collection__ = 'testdoc'
    s = fields.SetField(fields.IntegerField())


def test_default():
    doc = Doc()
    assert isinstance(doc.s, fields.set.Set)
    assert not doc.s
    assert len(doc.s) == 0
    assert doc.s == set()
    assert isinstance(doc.s._data, list)
    assert not doc.s._data


def test_get(db):
    _id = db.db.testdoc.insert({'s': [1, 2, 3]})
    doc = db.get_queryset(Doc).find_one(_id)

    assert doc.s
    assert len(doc.s) == 3
    assert doc.s._data == [1, 2, 3]
    assert doc.s == {1, 2, 3}

    with pytest.raises(TypeError):
        doc.s[1]

    with pytest.raises(TypeError):
        doc.s[1] = 10

    with pytest.raises(TypeError):
        del doc.s[1]


def test_add(db):
    _id = db.db.testdoc.insert({'s': [1, 2, 3]})
    doc = db.get_queryset(Doc).find_one(_id)
    doc.s.add(4)

    assert doc.s == {1, 2, 3, 4}


def test_add_typeerror():
    doc = Doc()

    with pytest.raises(ValueError):
        doc.s.add('not a number')


def test_add_save(db):
    _id = db.db.testdoc.insert({'s': [1, 2, 3]})
    doc = db.get_queryset(Doc).find_one(_id)
    doc.s.add(4)
    db.save(doc)

    data = db.db.testdoc.find_one({'_id': _id})
    assert data['s'] == [1, 2, 3, 4]


def test_remove(db):
    _id = db.db.testdoc.insert({'s': [1, 2, 3]})
    doc = db.get_queryset(Doc).find_one(_id)
    doc.s.remove(2)

    assert doc.s == {1, 3}


def test_remove_save(db):
    _id = db.db.testdoc.insert({'s': [1, 2, 3]})
    doc = db.get_queryset(Doc).find_one(_id)
    doc.s.remove(2)
    db.save(doc)

    data = db.db.testdoc.find_one({'_id': _id})
    assert data['s'] == [1, 3]


def test_add_to_set(db):
    _id = db.db.testdoc.insert({'s': [1, 2, 3]})
    doc = db.get_queryset(Doc).find_one(_id)
    doc.s.add_to_set(4)

    assert doc.s == {1, 2, 3, 4}

    data = db.db.testdoc.find_one({'_id': _id})
    assert data['s'] == [1, 2, 3, 4]


def test_add_to_set_typeerror():
    doc = Doc()
    with pytest.raises(ValueError):
        doc.s.add_to_set('not a number')


def test_pull(db):
    _id = db.db.testdoc.insert({'s': [1, 2, 3]})
    doc = db.get_queryset(Doc).find_one(_id)
    doc.s.pull(2)

    assert doc.s == {1, 3}

    data = db.db.testdoc.find_one({'_id': _id})
    assert data['s'] == [1, 3]



"""
This module for provide work with MongoDB database.

.. code-block:: python

    import pymongo
    from yadm.database import Database

    from mydocs import Doc

    client = pymongo.MongoClient("localhost", 27017)
    db = Database(self.client, 'test')

    doc = Doc()
    db.insert(doc)

    doc.arg = 13
    db.save(doc)

    qs = db.get_queryset(Doc).find({'arg': {'$gt': 10}})
    for doc in qs:
        print(doc)
"""
import pymongo

from yadm.markers import AttributeNotSet
from yadm.aggregation import Aggregator
from yadm.queryset import QuerySet
from yadm.bulk import Bulk
from yadm.serialize import to_mongo, from_mongo
from yadm.common import build_update_query


PYMONGO_VERSION = pymongo.version_tuple

RPS = pymongo.read_preferences


class BaseDatabase:
    aio = None

    def __init__(self, client, name, **database_params):
        self.client = client
        self.name = name
        self.database_params = database_params
        self.db = client.get_database(name, **database_params)

    def __repr__(self):  # pragma: no cover
        return '{}({!r})'.format(self.__class__.__name__, self.db)

    def __call__(self, document_class, **params):
        return self.get_queryset(document_class, **params)

    def _get_collection(self, document_class, params=None):
        """ Return pymongo collection for document class.
        """
        return self.db.get_collection(document_class.__collection__,
                                      **(params or {}))

    def insert(self, document, **collection_params):
        raise NotImplementedError

    def save(self, document, full=False, upsert=False, **collection_params):
        raise NotImplementedError

    def update_one(self, document, *, reload=True,
                   set=None, unset=None, inc=None,
                   push=None, pull=None,
                   **collection_params):
        raise NotImplementedError

    def remove(self, document, **collection_params):
        raise NotImplementedError

    def reload(self, document, new_instance=False, *,
               projection=None,
               **collection_params):
        raise NotImplementedError

    def get_queryset(self, document_class, *,
                     projection=None,
                     cache=None,
                     **collection_params):
        raise NotImplementedError

    def get_document(self, document_class, _id, *,
                     projection=None,
                     exc=None,
                     read_preference=RPS.PrimaryPreferred(),
                     **collection_params):
        raise NotImplementedError

    def aggregate(self, document_class, *,
                  pipeline=None, **collection_params):
        raise NotImplementedError

    def bulk(self, document_class, *,
             ordered=False, raise_on_errors=True, **collection_params):
        raise NotImplementedError


class Database(BaseDatabase):
    """ Main object who provide work with database.

    :param pymongo.Client client: database connection
    :param str name: database name
    """
    aio = False

    def insert(self, document, **collection_params):
        """ Insert document to database.

        :param Document document: document instance for insert to database

        It's bind new document to database set
        :py:attr:`_id <yadm.documents.Document._id>`.
        :param **collection_params: params for get_collection
        """
        document.__db__ = self
        collection = self._get_collection(document.__class__, collection_params)

        document._id = collection.insert_one(to_mongo(document)).inserted_id
        document.__changed_clear__()
        return document

    def save(self, document, full=False, upsert=False, **collection_params):
        """ Save document to database.

        :param Document document: document instance for save
        :param bool full: fully resave document
            (default: `False`)
        :param bool upsert: see documentation for MongoDB's `update`
            (default: `False`)
        :param **collection_params: params for get_collection

        If document has no `_id`
        :py:meth:`insert <Database.insert>` new document.
        """
        if hasattr(document, '_id'):
            document.__db__ = self

            if full:
                self._get_collection(document, collection_params).update(
                    {'_id': document.id},
                    to_mongo(document),
                    upsert=upsert,
                    multi=False,
                )
                document.__changed_clear__()
            else:
                set_data = to_mongo(
                    document,
                    exclude=['_id'],
                    include=list(document.__changed__),
                )

                unset_data = [f for f, v in document.__changed__.items()
                              if v is AttributeNotSet]

                self.update_one(document, set=set_data, unset=unset_data,
                                **collection_params)

            return document
        else:
            return self.insert(document, **collection_params)

    def update_one(self, document, *, reload=True,
                   set=None, unset=None, inc=None,
                   push=None, pull=None,
                   **collection_params):  # TODO: extend
        """ Update one document.

        :param Document document: document instance for update
        :param bool reload: if True, reload document
        :param **collection_params: params for get_collection
        """
        update_data = build_update_query(set=set, unset=unset, inc=inc,
                                         push=push, pull=pull)

        if update_data:
            self._get_collection(document, collection_params).update(
                {'_id': document.id},
                update_data,
                upsert=False,
                multi=False,
            )

        if reload:
            self.reload(document, **collection_params)

    def remove(self, document, **collection_params):
        """ Remove document from database.

        :param Document document: instance for remove from database
        :param **collection_params: params for get_collection
        """
        col = self._get_collection(document.__class__, collection_params)
        return col.remove({'_id': document._id})

    def reload(self, document, new_instance=False, *,
               projection=None,
               read_preference=RPS.PrimaryPreferred(),
               **collection_params):
        """ Reload document.

        :param Document document: instance for reload
        :param bool new_instance: if `True` return new instance of document,
            else change data in given document (default: `False`)
        :param dict projection: projection for query
        :param **collection_params: params for get_collection
        """
        collection_params['read_preference'] = read_preference
        qs = self.get_queryset(document.__class__,
                               projection=projection,
                               **collection_params)

        if projection is not None:
            new = qs.find_one(document.id, projection)
        else:
            new = qs.find_one(document.id)

        if new_instance:
            return new
        else:
            document.__raw__.clear()
            document.__raw__.update(new.__raw__)
            document.__cache__.clear()
            document.__changed__.clear()
            return document

    def get_document(self, document_class, _id, *,
                     projection=None,
                     exc=None,
                     read_preference=RPS.PrimaryPreferred(),
                     **collection_params):
        """ Get document for it _id.

        :param document_class: :class:`yadm.documents.Document`
        :param _id: document's _id
        :param dict projection: projection for query
        :param Exception exc: raise given exception if not found
        :param **collection_params: params for get_collection

        Default ReadPreference is PrimaryPreferred.
        """
        collection_params['read_preference'] = read_preference
        col = self.db.get_collection(document_class.__collection__,
                                     **collection_params)

        if projection is None:
            projection = document_class.__default_projection__

        if projection is not None:
            raw = col.find_one({'_id': _id}, projection)
            not_loaded = [k for k, v in projection.items() if not v]
        else:
            raw = col.find_one({'_id': _id})
            not_loaded = []

        if raw:
            doc = from_mongo(document_class, raw, not_loaded=not_loaded)
            doc.__db__ = self
            return doc

        elif exc is not None:
            raise exc((document_class, _id, collection_params))

        else:
            return None

    def get_queryset(self, document_class, *,
                     projection=None,
                     cache=None,
                     **collection_params):
        """ Return queryset for document class.

        :param document_class: :class:`yadm.documents.Document`
        :param dict projection:
        :param cache: cache for share with other querysets
        :param **collection_params: params for get_collection

        This create instance of :class:`yadm.queryset.QuerySet`
        with presetted document's collection information.
        """
        if projection is None:
            projection = document_class.__default_projection__

        return QuerySet(self, document_class,
                        projection=projection,
                        cache=cache,
                        collection_params=collection_params)

    def aggregate(self, document_class, *,
                  pipeline=None,
                  **collection_params):
        """ Return aggregator for use aggregation framework.

        :param document_class: :class:`yadm.documents.Document`
        :param list pipeline: initial pipeline
        :param **collection_params: params for get_collection
        """
        return Aggregator(self, document_class, pipeline=pipeline,
                          collection_params=collection_params)

    def bulk(self, document_class, *,
             ordered=False, raise_on_errors=True,
             **collection_params):
        """ Return Bulk.

        :param MetaDocument document_class: class of documents fo bulk
        :param bool ordered: create ordered bulk (default `False`)
        :param bool raise_on_errors: raise BulkWriteError exception
            if write errors (default `True`)
        :param **collection_params: params for get_collection

        Context manager:

            with db.bulk(Doc) as bulk:
                bulk.insert(doc_1)
                bulk.insert(doc_2)
        """
        return Bulk(self, document_class, ordered,
                    raise_on_errors, collection_params)



from enum import Enum

import pytest

from yadm import fields
from yadm.fields.enum import EnumStateSetError
from yadm.documents import Document


class WordsEnum(Enum):
    a = 1
    b = 2
    c = 3


class TestEnumField:
    class Doc(Document):
        __collection__ = 'docs'
        e = fields.EnumField(WordsEnum)

    def test_init(self):
        doc = self.Doc()
        assert not hasattr(doc, 'e')

    def test_save(self, db):
        doc = self.Doc()
        doc.e = WordsEnum.a
        assert doc.e == WordsEnum.a

        db.save(doc)
        raw = db.db['docs'].find_one()
        assert raw['e'] == 1

    def test_load(self, db):
        db.db['docs'].insert({'e': 2})
        doc = db(self.Doc).find_one()
        assert isinstance(doc.e, WordsEnum)
        assert doc.e == WordsEnum.b

    def test_set_valid_data(self):
        doc = self.Doc()
        doc.e = WordsEnum.a.value
        assert doc.e == WordsEnum.a

    def test_set_not_valid_data(self):
        doc = self.Doc()
        with pytest.raises(ValueError):
            doc.e = 'wrong'


class TestEnumStateField:
    class Doc(Document):
        __collection__ = 'docs'
        RULES = {
            WordsEnum.a: [WordsEnum.b],
            WordsEnum.b: [WordsEnum.c, WordsEnum.a],
        }
        e = fields.EnumStateField(WordsEnum, rules=RULES, start=WordsEnum.a)

    def test_init(self):
        doc = self.Doc()
        assert doc.e == WordsEnum.a

    def test_init__empty_rules(self):
        with pytest.raises(TypeError):
            fields.EnumField(WordsEnum, rules=[])

    def test_ok(self):
        doc = self.Doc()
        assert doc.e == WordsEnum.a

        doc.e = WordsEnum.b
        assert doc.e == WordsEnum.b

        doc.e = WordsEnum.c
        assert doc.e == WordsEnum.c

    def test_not_ok(self):
        doc = self.Doc()
        assert doc.e == WordsEnum.a

        with pytest.raises(EnumStateSetError):
            doc.e = WordsEnum.c

    def test_set_equal(self):
        doc = self.Doc()
        assert doc.e == WordsEnum.a

        doc.e = WordsEnum.a
        assert doc.e == WordsEnum.a



import pytest

from yadm.testing import create_fake
from yadm.documents import Document
from yadm import fields


class DocRef(Document):
    __collection__ = 'testdocs_ref'
    s = fields.StringField()


class Doc(Document):
    __collection__ = 'testdocs'
    ref = fields.ReferenceField(DocRef)
    i = fields.IntegerField()


def test_reference(db):
    doc = create_fake(Doc, db)
    assert isinstance(doc, Doc)
    assert isinstance(doc.ref, DocRef)
    assert doc.ref.id



""" Field for list with references.

Usage:

    class Doc(Document):
        refs = ReferencesListField(RefDoc)


    doc = db(Doc).find_one(...)
    doc.ref.resolve()  # resolve all documents in with query
    for ref_doc in doc.ref:
        ...

With asyncio:

    await doc.ref.resolve()

doc.ref is a ReferencesList instance, witch subclass MutableSequence.
But without resolving NotResolved raised for any actions with it
(except __len__ and __bool__).

"""
from collections.abc import MutableSequence

from yadm.documents import DocumentItemMixin
from yadm.queryset import NotFoundBehavior
from yadm.fields.base import Field


class NotResolved(Exception):
    pass


class AlreadyResolved(Exception):
    pass


class ReferencesList(MutableSequence, DocumentItemMixin):
    _resolved = False
    _changed = False

    def __init__(self, reference_document_class, ids: list,
                 field=None, parent=None):
        self._reference_document_class = reference_document_class
        self.__parent__ = parent
        self._field = field
        self._ids = ids or []
        self._documents = []

        if not ids:
            self._resolved = True

    def __repr__(self):
        if self._resolved:
            items = ', '.join([repr(d) for d in self._documents])
        else:
            items = ', '.join([str(i) for i in self._ids])

        if len(items) > 200:  # pragma: nocover
            items = items[:75] + ' ... ' + items[-75:]

        return "{cname}({rdc} {resolved}{changed} {len} [{items}])".format(
            cname=self.__class__.__name__,
            rdc=self._reference_document_class.__name__,
            resolved='r' if self.resolved else '',
            changed='c' if self.changed else '',
            len=len(self),
            items=items,
        )

    def __getitem__(self, idx: int):
        self._check_resolved_and_rise()
        return self._documents[idx]

    def __setitem__(self, idx: int, document: 'Document'):
        self._check_resolved_and_rise()
        self._set_changed()
        self._ids[idx] = document
        self._documents[idx] = document

    def __delitem__(self, idx: int):
        self._check_resolved_and_rise()
        self._set_changed()
        del self._ids[idx]
        del self._documents[idx]

    def __len__(self):
        return len(self._ids)

    def __bool__(self):
        return bool(self._ids)

    def __iter__(self):
        self._check_resolved_and_rise()
        return iter(self._documents)

    def __eq__(self, other):  # pragma: nocover
        if isinstance(other, ReferencesList):
            return self._ids == other._ids
        else:
            return NotImplemented

    @property
    def resolved(self) -> bool:
        return self._resolved

    @property
    def changed(self) -> bool:
        return self._changed

    @property
    def ids(self) -> list:
        return self._ids.copy()

    def insert(self, idx: int, document: 'Document'):
        self._check_resolved_and_rise()
        self._set_changed()
        self._ids.insert(idx, document.id)
        self._documents.insert(idx, document)

    def append(self, document: 'Document'):
        self._check_resolved_and_rise()
        self._set_changed()
        self._ids.append(document.id)
        self._documents.append(document)

    def pop(self, idx: int=-1):
        self._check_resolved_and_rise()
        self._set_changed()
        del self._ids[idx]
        return self._documents.pop(idx)

    def resolve(self):
        """ Resolve ids to documents.

        This method can be used with "await".
        """
        if self._resolved:
            raise AlreadyResolved()

        db = self.__db__
        qs = db.get_queryset(self._reference_document_class)

        if not db.aio:
            self._documents = list(qs.find_in(
                self._ids,
                not_found=NotFoundBehavior.NONE
            ))
            self._resolved = True
        else:
            async def resolver_coro(self):
                accumulator = []
                async for doc in qs.find_in(self._ids):
                    accumulator.append(doc)

                self._documents = accumulator
                self._resolved = True

            return resolver_coro(self)

    def _check_resolved_and_rise(self):
        if not self._resolved:
            raise NotResolved()

    def _set_changed(self):
        if not self._changed:
            self._changed = True

            # if self.__parent__ is not None:
            #     self._field.set_parent_changed(self)


class ReferencesListField(Field):
    def __init__(self, reference_document_class):
        self._reference_document_class = reference_document_class

    def copy(self):  # pragma: no cover
        return self.__class__()

    def get_if_attribute_not_set(self, document):  # pragma: no cover
        return ReferencesList(
            self._reference_document_class, [],
            field=self,
            parent=document,
        )

    def get_default(self, document):
        return ReferencesList(
            self._reference_document_class, [],
            field=self,
            parent=document,
        )

    def prepare_value(self, document, value):
        if isinstance(value, ReferencesList):
            value._field = self
            value.__parent__ = document
            self.set_parent_changed(value)
            return value

        elif isinstance(value, list):
            ids = [i.id for i in value]
            res = ReferencesList(
                self._reference_document_class, ids,
                field=self,
                parent=document,
            )
            res._resolved = True
            res._documents = value
            self.set_parent_changed(res)
            return res

        else:  # pragma: no cover
            raise TypeError(value)

    def from_mongo(self, document, value: list):
        return ReferencesList(
            self._reference_document_class, value,
            field=self,
            parent=document,
        )

    def to_mongo(self, document, value: ReferencesList):
        return value._ids



from yadm.fields.base import pass_null
from yadm.fields.simple import SimpleField
from yadm.markers import AttributeNotSet


class EnumField(SimpleField):
    """ Field for enum.Enum .
    """
    def __init__(self, enum, **kwargs):
        self.type = enum
        super().__init__(**kwargs)

    @pass_null
    def to_mongo(self, document, value):
        return value.value


class EnumStateSetError(Exception):
    def __init__(self, current, new):
        self.current = current
        self.new = new
        self.message = ("Not allowed in rules: {} -> {}"
                        "".format(current, new))

        super().__init__(self.message)


class EnumStateField(EnumField):
    """ Simple state machine with states are enum.Enum .
    """
    rules = None

    def __init__(self, enum, rules=None, start=AttributeNotSet, **kwargs):
        super().__init__(enum, default=start, **kwargs)

        if rules is not None:
            self.rules = rules

        if not self.rules:
            raise ValueError("Rules list is empty")

    @pass_null
    def prepare_value(self, document, value):
        current_value = getattr(document, self.name, AttributeNotSet)
        new_value = super().prepare_value(document, value)

        if (new_value != current_value and
                new_value not in self.rules.get(current_value, [])):
            raise EnumStateSetError(current_value, new_value)

        return new_value



import pytest

from bson import ObjectId

from yadm import fields
from yadm.documents import Document


class Doc(Document):
    __collection__ = 'testdoc'
    map = fields.MapField(fields.IntegerField())


def test_default():
    doc = Doc()
    assert isinstance(doc.map, fields.map.Map)
    assert not doc.map
    assert len(doc.map) == 0
    assert doc.map._data == {}
    assert doc.map == {}


def test_load(db):
    _id = db.db.testdoc.insert({'map': {'a': 1, 'b': 2, 'c': 3}})
    doc = db.get_queryset(Doc).find_one(_id)

    assert doc.map
    assert len(doc.map) == 3
    assert doc.map._data == {'a': 1, 'b': 2, 'c': 3}
    assert dict(doc.map) == {'a': 1, 'b': 2, 'c': 3}
    assert doc.map == {'a': 1, 'b': 2, 'c': 3}
    assert doc.map['b'] == 2


def test_getitem_keyerror():
    doc = Doc()
    with pytest.raises(KeyError):
        doc.map['not exist']


def test_setitem(db):
    _id = db.db.testdoc.insert({'map': {'a': 1, 'b': 2, 'c': 3}})
    doc = db.get_queryset(Doc).find_one(_id)
    doc.map['d'] = 4

    assert doc.map == {'a': 1, 'b': 2, 'c': 3, 'd': 4}


def test_setitem_valueerror():
    doc = Doc()
    with pytest.raises(ValueError):
        doc.map['d'] = 'not a number'


def test_setattr_save(db):
    _id = db.db.testdoc.insert({'map': {'a': 1, 'b': 2, 'c': 3}})
    doc = db.get_queryset(Doc).find_one(_id)
    doc.map['d'] = 4
    db.save(doc)

    data = db.db.testdoc.find_one({'_id': _id})
    assert data['map'] == {'a': 1, 'b': 2, 'c': 3, 'd': 4}
    assert doc.map == {'a': 1, 'b': 2, 'c': 3, 'd': 4}


def test_remove(db):
    _id = db.db.testdoc.insert({'map': {'a': 1, 'b': 2, 'c': 3}})
    doc = db.get_queryset(Doc).find_one(_id)
    del doc.map['b']

    assert doc.map == {'a': 1, 'c': 3}


def test_remove_save(db):
    _id = db.db.testdoc.insert({'map': {'a': 1, 'b': 2, 'c': 3}})
    doc = db.get_queryset(Doc).find_one(_id)
    del doc.map['b']
    db.save(doc)

    data = db.db.testdoc.find_one({'_id': _id})
    assert data['map'] == {'a': 1, 'c': 3}


def test_set(db):
    _id = db.db.testdoc.insert({'map': {'a': 1, 'b': 2, 'c': 3}})
    doc = db.get_queryset(Doc).find_one(_id)
    doc.map.set('d', 4)

    assert doc.map == {'a': 1, 'b': 2, 'c': 3, 'd': 4}

    data = db.db.testdoc.find_one({'_id': _id})
    assert data['map'] == {'a': 1, 'b': 2, 'c': 3, 'd': 4}


def test_set_runtimeerror():
    doc = Doc()
    with pytest.raises(RuntimeError):
        doc.map.set('key', 1)


def test_set_valueerror(db):
    doc = Doc()
    db.save(doc)
    with pytest.raises(ValueError):
        doc.map.set('key', 'not a number')


def test_unset(db):
    _id = db.db.testdoc.insert({'map': {'a': 1, 'b': 2, 'c': 3}})
    doc = db.get_queryset(Doc).find_one(_id)
    doc.map.unset('b')

    assert doc.map == {'a': 1, 'c': 3}

    data = db.db.testdoc.find_one({'_id': _id})
    assert data['map'] == {'a': 1, 'c': 3}


class DocCustom(Document):
    __collection__ = 'testdoc'
    map = fields.MapCustomKeysField(fields.IntegerField(), ObjectId)


def test_custom_load(db):
    one = ObjectId()
    two = ObjectId()

    _id = db.db.testdoc.insert({'map': {str(one): 1, str(two): 2}})
    doc = db.get_queryset(DocCustom).find_one(_id)

    assert doc.map
    assert len(doc.map) == 2
    assert doc.map._data == {str(one): 1, str(two): 2}
    assert dict(doc.map) == {one: 1, two: 2}
    assert doc.map == {one: 1, two: 2}
    assert doc.map[two] == 2
    assert isinstance(list(doc.map)[0], ObjectId)


def test_custom_setitem(db):
    one = ObjectId()
    two = ObjectId()

    _id = db.db.testdoc.insert({'map': {str(one): 1}})
    doc = db.get_queryset(DocCustom).find_one(_id)

    doc.map[two] = 3
    assert doc.map == {one: 1, two: 3}

    doc.map[two] = 2
    assert doc.map == {one: 1, two: 2}


def test_complex_references(db):
    class RefDoc(Document):
        __collection__ = 'refs'
        i = fields.IntegerField()

    class Doc(Document):
        __collection__ = 'docs'
        refs = fields.MapCustomKeysField(
            fields.ListField(fields.ReferenceField(RefDoc)),
            key_factory=ObjectId)

    _id = ObjectId()
    ref = db.save(RefDoc(i=1))
    doc = db.save(Doc(refs={_id: [ref]}))
    assert doc.refs == {_id: [ref]}

    doc = db(Doc).find_one({'_id': doc.id})
    assert doc.refs == {_id: [ref]}
    assert doc.refs[_id][0] is not ref



"""
Functions for serialize and deserialize data.
"""

from collections import defaultdict

from yadm.markers import AttributeNotSet, NotLoaded


def _to_mongo_value(result, document, field, name, value):
    if value is not AttributeNotSet:
        result[name] = field.to_mongo(document, value)


def to_mongo(document, exclude=(), include=None):
    """ Serialize document to MongoDB data.

    :param BaseDocument document: document for serializing
    :param list exclude: exclude fields
    :param list include: include only fields (all by default)
    """
    result = {}

    for name, field in document.__fields__.items():
        if name in exclude:
            continue

        elif include is not None and name not in include:
            continue

        elif name in document.__changed__:
            value = document.__changed__[name]
            _to_mongo_value(result, document, field, name, value)

        elif name in document.__raw__:
            value = document.__raw__[name]
            if value is AttributeNotSet:
                continue

            result[name] = document.__raw__[name]

        elif name in document.__cache__:
            value = document.__cache__[name]
            _to_mongo_value(result, document, field, name, value)

        else:
            value = field.get_default(document)
            document.__changed__[name] = value  # lazy load default value
            _to_mongo_value(result, document, field, name, value)

    if include:
        include_groups = defaultdict(set)

        for name in (f for f in include if '.' in f):
            first, last = name.split('.', 1)
            include_groups[first].add(last)

        for name, subinclude in include_groups.items():
            edoc = getattr(document, name)
            result[name] = to_mongo(edoc, include=subinclude)

    return result


def from_mongo(document_class, data, not_loaded=(), parent=None, name=None):
    """ Deserialize MongoDB data to document.

    :param document_class: document class
    :param dict data: data from MongoDB
    :param list not_loaded: fields,
        who marked as not loaded
    :param parent: parent for new document
    :param str name: name for new document
    """
    document = document_class()

    for field_name, field in document_class.__fields__.items():
        if field_name in data:
            document.__raw__[field_name] = data[field_name]

        elif field_name in not_loaded:
            document.__raw__[field_name] = NotLoaded

        elif field.smart_null:
            document.__raw__[field_name] = None

        else:
            document.__raw__[field_name] = AttributeNotSet

    if parent is not None:
        document.__parent__ = parent
        document.__name__ = name

    return document



import abc


class CacheInterface(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __getitem__(self, key):  # pragma: no cover
        pass

    @abc.abstractmethod
    def __setitem__(self, key, value):  # pragma: no cover
        pass

    @abc.abstractmethod
    def __delitem__(self, key):  # pragma: no cover
        pass

    @abc.abstractmethod
    def __contains__(self, key):  # pragma: no cover
        pass

    @abc.abstractmethod
    def __len__(self, key):  # pragma: no cover
        pass


@CacheInterface.register
class NotACache:
    def __getitem__(self, key):  # pragma: no cover
        raise KeyError(key)

    def __setitem__(self, key, value):  # pragma: no cover
        pass

    def __delitem__(self, key):  # pragma: no cover
        pass

    def __contains__(self, key):  # pragma: no cover
        return False

    def __len__(self, key):  # pragma: no cover
        return 0


@CacheInterface.register
class StackCache(dict):
    def __init__(self, size):
        self.size = size
        self._stack = []

    def __setitem__(self, key, item):
        if key not in self:
            self._stack.append(key)

            if len(self._stack) > self.size:
                del self[self._stack[0]]

        super().__setitem__(key, item)

    def __delitem__(self, key):
        super().__delitem__(key)
        self._stack.remove(key)



import pytest

from yadm import fields
from yadm.documents import Document
from yadm.testing import create_fake


class TestStaticField:
    class Doc(Document):
        __collection__ = 'docs'

        int = fields.StaticField(13)
        str = fields.StaticField('string')

    def test_init(self):
        doc = self.Doc()

        assert doc.int == 13
        assert doc.str == 'string'

    def test_set(self):
        doc = self.Doc()

        with pytest.raises(AttributeError):
            doc.int = 26

        with pytest.raises(AttributeError):
            doc.int = 13

        with pytest.raises(AttributeError):
            doc.str = ''

        with pytest.raises(AttributeError):
            doc.str = None

    def test_save(self, db):
        doc = self.Doc()
        db.save(doc)

        raw = db.db['docs'].find_one()

        assert raw['int'] == 13
        assert raw['str'] == 'string'

    def test_load(self, db):
        db.db['docs'].insert({'int': 13, 'str': 'string'})

        doc = db(self.Doc).find_one()

        assert doc.int == 13
        assert doc.str == 'string'

    @pytest.mark.parametrize('data, field', [
        ({'int': 26, 'str': 'string'}, 'int'),
        ({'int': 13, 'str': 'bad string'}, 'str'),
        ({'int': None, 'str': 'string'}, 'int'),
        ({'str': 'string'}, 'int'),
    ])
    def test_bad_data_in_db(self, db, data, field):
        db.db['docs'].insert(data)

        doc = db(self.Doc).find_one()

        with pytest.raises(RuntimeError):
            getattr(doc, field)

    def test_faker(self):
        doc = create_fake(self.Doc)
        assert doc.int == 13
        assert doc.str == 'string'

        with pytest.raises(AttributeError):
            doc.int = 666

    def test_faker_save(self, db):
        doc = create_fake(self.Doc, db)
        assert doc.int == 13
        assert doc.str == 'string'

        raw = db.db[self.Doc.__collection__].find_one()
        assert raw['int'] == 13
        assert raw['str'] == 'string'



import pytest

from bson import ObjectId

from yadm.documents import Document, EmbeddedDocument
from yadm import fields


class DocRef(Document):
    __collection__ = 'testdocs_ref'


class Doc(Document):
    __collection__ = 'testdocs'
    ref = fields.ReferenceField(DocRef)


def test_get(db):
    id_ref = db.db.testdocs_ref.insert({})
    id = db.db.testdocs.insert({'ref': id_ref})

    doc = db.get_queryset(Doc).find_one(id)

    assert doc._id == id
    assert isinstance(doc.ref, Document)
    assert doc.ref._id == id_ref


def test_get_broken(db):
    id = db.db.testdocs.insert({'ref': ObjectId()})
    doc = db.get_queryset(Doc).find_one(id)

    with pytest.raises(fields.BrokenReference):
        doc.ref


def test_get_notbindingtodatabase(db):
    id = db.db.testdocs.insert({'ref': ObjectId()})
    doc = db.get_queryset(Doc).find_one(id)
    doc.__db__ = None

    with pytest.raises(fields.NotBindingToDatabase):
        doc.ref


@pytest.mark.parametrize('cast', [
    lambda d: d,
    lambda d: d.id,
    lambda d: str(d.id),
])
def test_set(db, cast):
    id = db.db.testdocs.insert({})
    doc = db.get_queryset(Doc).find_one(id)
    doc.ref = cast(db.insert(DocRef()))

    assert isinstance(doc.ref, DocRef)

    db.save(doc)

    id_ref = db.db.testdocs.find_one({'_id': id})['ref']

    assert isinstance(id_ref, ObjectId)
    assert isinstance(doc.ref, Document)
    assert id_ref == doc.ref.id


def test_not_object_id_get(db):
    class DocNotObjectIdRef(Document):
        __collection__ = 'testdocs_ref'
        _id = fields.IntegerField()

    class DocNotObjectId(Document):
        __collection__ = 'testdocs'
        ref = fields.ReferenceField(DocNotObjectIdRef)

    id_ref = db.db.testdocs_ref.insert({'_id': 13})
    id = db.db.testdocs.insert({'ref': id_ref})

    doc = db.get_queryset(DocNotObjectId).find_one(id)

    assert doc._id == id
    assert isinstance(doc.ref, Document)
    assert doc.ref._id == id_ref


class DocEmb(EmbeddedDocument):
    ref = fields.ReferenceField(DocRef)


class DocWEmbedded(Document):
    __collection__ = 'testdocs'
    emb = fields.EmbeddedDocumentField(DocEmb)


def test_embedded_get(db):
    id_ref = db.db.testdocs_ref.insert({})
    id = db.db.testdocs.insert({'emb': {'ref': id_ref}})

    doc = db.get_queryset(DocWEmbedded).find_one(id)

    assert doc._id == id
    assert isinstance(doc.emb, DocEmb)
    assert isinstance(doc.emb.ref, DocRef)
    assert doc.emb.ref._id == id_ref


def test_cache(db):
    ref_one = db.insert(DocRef(i=13))
    ref_two = db.insert(DocRef(i=26))

    db.insert(Doc(ref=ref_one))
    db.insert(Doc(ref=ref_one))
    db.insert(Doc(ref=ref_two))

    qs = db.get_queryset(Doc)
    assert len(qs.cache) == 0
    assert len(qs) == 3
    assert len({d.ref for d in qs}) == 2
    assert len({id(d.ref) for d in qs}) == 2
    assert len(qs.cache) == 2

    doc = qs.find_one()
    assert doc.ref.__qs__ is not doc.__qs__
    assert doc.ref.__qs__.cache is doc.__qs__.cache



import pytest

from yadm.documents import Document
from yadm.fields.email import EmailField, InvalidEmail


class Doc(Document):
    e = EmailField()


def test_ok():
    doc = Doc()
    doc.e = 'E@mA.iL'

    assert doc.e == 'e@ma.il'


def test_error():
    doc = Doc()

    with pytest.raises(InvalidEmail):
        doc.e = 'EmA.iL'
